#include <stdio.h>
#include <stdlib.h>
#include <string.h>      // memset
#include <sys/socket.h>
#include <sys/ioctl.h>   // for number of interfaces
#include <net/if.h>      // for ifreq
#include <netinet/in.h>  // for sockaddr_in
#include <netinet/tcp.h> // for tcp_info
#include <unistd.h>
#include <math.h>        // for floor

#include "common_meas.h" // common functions/structs
#include "common_meas.c" // common functions/structs

#define TXTYPE_FILE   0x1
#define TXTYPE_STRING 0x2

#pragma pack(1)         // this avoids padding in structs, it works (at least)
                        //  in mips architecture

struct ifreq ifreq[MAXINTERFACES];

int set_mptcp(int value);
int establish(int srcport, char * dstipa, int dstport, int nbr_if, double* conn_time, int sock_ex);
//int establish(int srcport, char * dstipa, int dstport, int nbr_if, double* conn_time);
int send_indication(int sock_control, int experiment);
int send_filechar(int sock_control, int txtype, int stortype, char* string, char* suffix);
struct nbr_port* receive_dataport(int sock);
void send_result(int sock_control, double bw_benefit, double delay_benefit);
//int test_send_cbr(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric);
int test_send_cbr(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric, int sock_ex);
int test_recv(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname);
int test_recv_unif(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric);
int getsock_tcp_mss(int inSock);
void send_abort(int sock_control);

unsigned char id_client;

/* Command-line parameters */

struct global_var_client gl_def;
struct global_var_client gl;

struct timeval t0, now;

int srcdataport = 0; // no need of fixing source ports

int main(int argc, char** argv)
{
//  fprintf(stdout, "u_long %ld, u_int %ld, u_short %ld, u_char %ld\nlong %ld, int %ld double %ld\n", sizeof(u_long), sizeof(u_int), sizeof(u_short), sizeof(u_char), sizeof(long), sizeof(int), sizeof(double));


  int b[MAXINTERFACES + 1];       // List of interfaces

  int a[MAXINTERFACES + 1];       // Sequence of tests to be performed
                                  // Semantics :
                                  //     a[i] < interfaces      ==> TCP connection ONLY on interface a[i]
                                  //     a[i] == interfaces + 1 ==> MPTCP connection on all interfaces              ...as it is now
                                  //     a[i] == interfaces + 2 ==> simultaneous TCP connections on ALL interfaces
  int len_a = 0;                      // Length (effective) for array a. av_ifs+1 if metric != BW2, 2 if metric == BW2
  int vret[MAXINTERFACES + 1];

  char* f_name1 = "output-client-test1.txt", *f_name2 = "output-client-test2.txt";
  char order[BUFFSIZE], order2[BUFFSIZE], order3[BUFFSIZE];
  char tsaf[BUFFSIZE], addstr[BUFFSIZE], addstr2[BUFFSIZE], addstr2b[BUFFSIZE];
  char initial[BUFFSIZE], outpserver[BUFFSIZE], addstr3[BUFFSIZE], addstr3b[BUFFSIZE];
  char outpserver2[BUFFSIZE]; // log for tcp conn estab times

  int i, j, k, numbindtcp;
  int old_mp_enabled;
  char popout = 0;

  int doublelocal, nlo, totret;

  //struct measure w_only_tcp[MAXINTERFACES], w_mptcp;
  struct measure * w_only_tcp, * w_sim_tcp, * w_mptcp;
  double max_bw = 0, max_bw_sim = 0, tot_bw = 0, tot_bw_sim = 0;
  double min_delay = 99999, max_delay = 0;
  double min_delay2 = 99999, max_delay2 = 0;
  int ret=0;

  int sock, sock_control, dataport;
  struct ifconf ifconf;
  int interfaces, av_ifs;

  pid_t pid;
  pid_t vpid[MAXINTERFACES]; // for tcpdump processes on MPTCP transmission

  double bw_benefit, delay_benefit, delay2_benefit, tdi;

  /* For debug shuffle function (can be removed)
  av_ifs=2;
  a[0]=1; a[1]=2; a[2]=4;
  shuffle(a, av_ifs+1);

  printf("%d %d %d (RAND_MAX %d)\n", a[0], a[1], a[2], RAND_MAX);
  exit(0);
  /*

  /* Preliminary checks */

  if (checkroot() != 0) {
    printf("You must have root privileges to run this client.\n");
    exit(1);
  }

  /* Default values */

  gl_def.verbose = 0;
  gl_def.to_transmit = 3500000;          // to transmit or to receive (depending on the test)
  gl_def.slices = 1;
  gl_def.local = 0;                      // for debugging in local machine (and loopback interface)
  gl_def.debug = 0;                      // Debug mode (additional notifications)
  gl_def.use6 = 0;
  gl_def.port = 993;                     // Default value, 993
  strcpy(gl_def.ipa, "130.104.229.234"); // Default value, IPv4 from UCL server
  strcpy(gl_def.test, "oneway");         
  strcpy(gl_def.tag, "notag");
  gl_def.collect_traces = 0;
  gl_def.ntest = ONEWAY;
  gl_def.nmetric = DELAY;
  strcpy(gl_def.metric, avmetrics[gl_def.nmetric]);
  gl_def.secmin = 60;                     // Configurable in command line (min reasonable value, 60)
  gl_def.cbrate = 250000;                 // 250kbps

  /* Arguments parsing */ 

  parse_arguments(DEV_CLIENT, argc, argv, &gl_def, &gl);

  w_only_tcp = (struct measure *)malloc(MAXINTERFACES*(gl.slices+1)*sizeof(struct measure));
  w_sim_tcp = (struct measure *)malloc(MAXINTERFACES*(gl.slices+1)*sizeof(struct measure));

  w_mptcp = (struct measure *)malloc((gl.slices+1)*sizeof(struct measure));

  if (w_only_tcp == NULL || w_mptcp == NULL || w_sim_tcp == NULL)
  {
    fprintf(stdout, "ERROR, unable to malloc w_only_tcp, w_sim_tcp and w_mptcp structs\n");
    exit(1);
  }

  if (gl.verbose)
  {
    fprintf(stdout, "\n"); // Separate from rest of the shell / previous calls

//    fprintf(stdout, "Destination IPA: %s ; Destination port: %dMetric: %s ; Test: %s\n", gl.ipa, gl.port, (gl.nmetric == BW ? "bw" : "delay" ), gl.test);
    fprintf(stdout, "Destination IPA: %s ; Destination port: %d ; Metric: %s ; Test: %s\n", gl.ipa, gl.port, gl.metric, gl.test);

    if (gl.nmetric == BW || gl.nmetric == BW2 || gl.nmetric == BWALL || gl.nmetric == CBRTIME)
    {
      fprintf(stdout, "Sending time: %u sec ; Number of slices: %d\n", gl.secmin, gl.slices);
    }
    else if (gl.nmetric == DELAY || gl.nmetric == CBRBYTES)
    {
      fprintf(stdout, "To transmit: %u bytes ; Number of slices: %d\n", gl.to_transmit, gl.slices);      
    }
  }

  /* Preliminary checks */

  // 0. id_client file

#ifdef IDCLIENT
  id_client = IDCLIENT;
#else
  FILE *fpid = fopen("./id_client", "r");
  if (fpid == NULL)
  {
    id_client = 0xff;
    fprintf(stdout, "Warning ! id_client not found, using value %d as client id instead.\n", (int) id_client); 
  }
  else
  {
    fscanf(fpid, "%c", &id_client);
    id_client = (char) atoi(&id_client);
    fclose(fpid);
  }
#endif

  if (gl.verbose) fprintf(stdout, "Client id : %d\n", id_client);

/* Rather: force (-f) overwritting when running gzip
  // 1. If trace-collecting enabled, remove all PCAP/PCAP.GZ files, if any

  if (gl.collect_traces)
  {
    // To prevent further gzip errors...
    if ((ret = system("rm 0000000000-*.pcap*")) < 0 )
    { 
      fprintf(stdout, "Error when removing 0000000000-*.pcap* files.\n");
      exit(1);
    }
  }
*/

  // 2. Detection of available interfaces
 
  // Create a socket or return an error.
  if (!gl.use6)
    sock = socket(AF_INET, SOCK_STREAM, 0);
  else
    sock = socket(AF_INET6, SOCK_STREAM, 0);
  if (sock < 0)
  {
    fprintf(stdout, "Error in creating socket.\n");
    exit(1);
  }
 
  // Point ifconf's ifc_buf to our array of interface ifreqs.
  ifconf.ifc_buf = (char *) ifreq;
    
  // Set ifconf's ifc_len to the length of our array of interface ifreqs.
  ifconf.ifc_len = sizeof ifreq;
 
  //  Populate ifconf.ifc_buf (ifreq) with a list of interface names and flags.
  if (ioctl(sock, SIOCGIFCONF, &ifconf) == -1)
  {
    fprintf(stdout, "ioctl error when retrieving configuration.\n");
    exit(1);
  }
 
  // Divide the length of the interface list by the size of each entry.
  interfaces = ifconf.ifc_len / sizeof(ifreq[0]);

  // interfaces = number of interfaces (loopback and down interfaces included)
  // av_ifs     = number of available interfaces (<= interfaces) (only up interfaces, loopback excluded) 

  for (i=0, av_ifs=0; i<interfaces; i++)
  {

    if (ioctl(sock, SIOCGIFFLAGS, &ifreq[i]) == -1)
    {
      fprintf(stdout, "ioctl error when retrieving flags.\n");
      exit(1);
    }

    // Exclude loopback in normal operation (except for -l option)
    // and exclude interfaces not in UP state
    if ( ( (! (ifreq[i].ifr_flags & IFF_LOOPBACK)) || gl.local ) && (ifreq[i].ifr_flags & IFF_UP) )
    {
      if (gl.nmetric == BW || gl.nmetric == DELAY || gl.nmetric == CBRTIME || gl.nmetric == CBRBYTES || gl.nmetric == BWALL) // everything except BW2
      {
        a[av_ifs]=i;
        b[av_ifs]=i;
      }
      else if (gl.nmetric == BW2)
        b[av_ifs]=i;
      av_ifs ++;
      // if (gl.verbose) fprintf(stdout, "%s:%d ", ifreq[i].ifr_name, ifreq[i].ifr_flags);
    }
  }
  if (gl.verbose) fprintf(stdout, "\n");
  if (gl.nmetric == BW || gl.nmetric == DELAY || gl.nmetric == CBRTIME || gl.nmetric == CBRBYTES) // everything except BW2
  {
   a[av_ifs] = interfaces+1;      // Meaning: MPTCP connection over all interfaces
   len_a = av_ifs + 1;
  }

  if (gl.nmetric == BW2) { 
    a[0] = interfaces + 1;         // Meaning: MPTCP connection over all interfaces
    a[1] = interfaces + 2;         // Meaning: Simultaneous TCP connections over all interfaces

    len_a = 2;
  }

  if (gl.nmetric == BWALL) { 
    a[av_ifs] = interfaces + 1;      // Meaning: MPTCP connection over all interfaces
    a[av_ifs + 1] = interfaces + 2;  // Meaning: Simultaneous TCP connections over all interfaces

    len_a = av_ifs + 2;
  }

  gl.avifs = av_ifs;  // ok, redundant with av_ifs... but allows to use it everywhere without changing stuff 211114

  if (gl.verbose) fprintf(stdout, "Number of interfaces: %2d ; Number of available interfaces: %2d\n", interfaces, av_ifs);

  if (close(sock) != 0)
  {
    printf("ERROR, socket for interface discovery could not be successfully closed.\n");
  } 

  // 3. Is MPTCP supported ?
  if ((i = system("sysctl -n net.mptcp.mptcp_enabled > /dev/null")) < 0 )
  { 
    fprintf(stdout, "MPTCP is not supported in this machine.\n");
    exit(1);
  }

  // 4. Save old value of mptcp_enabled
  FILE *pp = popen("/sbin/sysctl -n net.mptcp.mptcp_enabled", "r");
  fscanf(pp, "%c\n", &popout);
  old_mp_enabled = atoi(&popout);

  // Why do you care on whether mptcp is enabled or not for the control channel ?
/*
  // 2b. Set mptcp to 0 (for control channel)
  set_mptcp(0); // It does not return anything, just exits if error
*/

  // 5. Disable tcp metrics
  memset(order, 0, BUFFSIZE); 
  sprintf(order, "sysctl -w net.ipv4.tcp_no_metrics_save=1 > /dev/null");
  if ((i = system(order)) < 0 )
  { 
    fprintf(stdout, "Error when enabling tcp_no_metrics_save in sysctl.\n");
    exit(1);
  }

  pclose(pp);

  /* Test call */

  get_now( &t0, 0 );

  /* Note : sysctl call cannot be used because no support for MPTCP 
     is provided in linux/sysctl.h for now */

//  sock_control = establish(0, gl.ipa, gl.port, NOINTERFACE, NULL);
  sock_control = establish(0, gl.ipa, gl.port, NOINTERFACE, NULL, 0);

  // 1. Randomize the array for test order

  //for (i=0; i<len_a; i++)
  //{
  //  printf("a[%d]=%d\n", i, a[i]);
  //}

  shuffle(a, len_a);
  if (gl.nmetric == BW2 || gl.nmetric == BWALL) shuffle(b, av_ifs);

  //printf("After shuffling...\n");
  //for (i=0; i<len_a; i++)
  //{
  //  printf("a[%d]=%d\n", i, a[i]);
  //}

  // 2. Run tests in randomized order

  if (gl.verbose)
  {
    fprintf(stdout, "Test order: ");

    // for(i = 0; i < av_ifs + 1; i++) // replaced "interfaces" with "av_ifs"
    for(i = 0; i < len_a; i++) // replaced "interfaces" with "av_ifs"
    {
      fprintf(stdout, "%d ", a[i]);
    }
    fprintf(stdout, "\n");
  }

//------------------------ new 31/10/2014
  int pid_sc[MAXINTERFACES];
  int status, sock_sc[MAXINTERFACES];
  int pipefd[MAXINTERFACES][2];
  u_char* buff;

//------------------------ end new

//  for (i = 0, j = 0; i < av_ifs + 1; i ++) // replaced "interfaces" with "av_ifs"
  for (i = 0, numbindtcp = 0; i < len_a; i ++)
  {

    // i: index of "a", sequence of commands
    // numbindtcp: increased at the end of the iteration of individual TCP measurement

    if (gl.verbose) fprintf(stdout, "\nTEST NUMBER %d : ", a[i]);

    if (a[i] < interfaces + 1) // without mptcp         // TO CHECK: should not be av_ifs ? which relationship ?
    {
      // 2.1. Without MPTCP
      //      (tests over all non-loopback interfaces)

      if (gl.verbose) fprintf(stdout, "Without MPTCP, interface %d (%s)\n", a[i], ifreq[a[i]].ifr_name);
      // This should be a function, it is a piece of code that appears several times
      if (gl.local)
      {
        if ( ! (ifreq[a[i]].ifr_flags & IFF_LOOPBACK) ) 
        {
          if (gl.verbose) fprintf(stdout, "Not loopback in mode 'local': leaving...\n");
          continue; // provisional: only use lo interface
        }
      }
      else
      {
        if ((ifreq[a[i]].ifr_flags & IFF_LOOPBACK) || ! (ifreq[a[i]].ifr_flags & IFF_UP)) continue;
      }

      // this (upper) does not need to be adapted to IPv6 ?

      //printf("interface %d\n",i);

      switch(gl.ntest)
      {
        case ONEWAY:
          dataport = send_indication(sock_control, ONEWAY);

          if (gl.collect_traces) pid = launch_tcpdump(ifreq[a[i]].ifr_name, DEV_CLIENT, (int) id_client, 0, i, dataport, order);

//          if ( ( ret = ( gl.nmetric == BW ? test_send_time(&(w_only_tcp[numbindtcp*(gl.slices)]), gl.ipa, dataport, a[i], 0, f_name1) : test_send(&(w_only_tcp[numbindtcp*(gl.slices)]), gl.ipa, dataport, a[i], 0, f_name1) )) < 0 )
//          if ( ( ret = test_send_cbr(&(w_only_tcp[numbindtcp*(gl.slices)]), gl.ipa, dataport, a[i], 0, f_name1, gl.nmetric) ) < 0 )
          if ( ( ret = test_send_cbr(&(w_only_tcp[numbindtcp*(gl.slices)]), gl.ipa, dataport, a[i], 0, f_name1, gl.nmetric, 0) ) < 0 )
          {
            fprintf(stdout, "Test with regular TCP (over %s) failed! return value = %d\n", ifreq[a[i]].ifr_name, ret);
            exit(1);
          }

          if (gl.collect_traces)
          {
            sleep(1); // give time to socket closing instructions to complete

            ret = stop_tcpdump(pid, DEV_CLIENT, order); // if ret!=0, something went wrong, don't send traces

            strcpy(order3, order);                    // Save PCAP filename in order3
            concatenate_string(order, ".gz");         // order contains PCAP.GZ filename

            // Send trace

            sprintf(order2, "client-%1d.pcap.gz", i);  // limited to 32 (size of suffix)  

            if( ret == 0 ) 
            {
              // file exists (everything ok)
              send_filechar(sock_control, TXTYPE_FILE, CLI_FILE, order, order2);

              sprintf(order2, "rm %s", order);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting compressed trace file.\n");
                exit(1);
              }
            } 
            else 
            {
              fprintf(stdout, "Non-fatal error (ret=%d), something went wrong in tcpdump/gzip, no tracefile will be sent to the server.\n", ret);
              
              // remove PCAP and/or PCAP.GZ file (whatever exists)
              sprintf(order2, "rm %s*", order3);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting trace file(s).\n");
                exit(1);
              }
            }
          }

          break;
        case REQRESP:

          dataport = send_indication(sock_control, REQRESP);

          if (gl.collect_traces) pid = launch_tcpdump(ifreq[a[i]].ifr_name, DEV_CLIENT, (int) id_client, 0, i, dataport, order);

          if ( ( ret = test_recv_unif(&(w_only_tcp[numbindtcp*(gl.slices)]), gl.ipa, dataport, a[i], 0, f_name1, gl.nmetric) ) < 0 )
          {
            fprintf(stdout, "Test with regular TCP (over %s) failed! return value = %d\n", ifreq[a[i]].ifr_name, ret);
            exit(1);
          }

          if (gl.collect_traces)
          {
            sleep(1); // give time to socket closing instructions to complete

            ret = stop_tcpdump(pid, DEV_CLIENT, order); // if ret!=0, something went wrong, don't send traces

            strcpy(order3, order);                    // Save PCAP filename in order3
            concatenate_string(order, ".gz");         // order contains PCAP.GZ filename

            // Send trace

            sprintf(order2, "client-%1d.pcap.gz", i);  // limited to 32 (size of suffix)  

            if( ret == 0 ) 
            {
              // file exists (everything ok)
              send_filechar(sock_control, TXTYPE_FILE, CLI_FILE, order, order2);

              sprintf(order2, "rm %s", order);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting compressed trace file.\n");
                exit(1);
              }
            } 
            else 
            {
              fprintf(stdout, "Non-fatal error (ret=%d), something went wrong in tcpdump/gzip, no tracefile will be sent to the server.\n", ret);
              
              // remove PCAP and/or PCAP.GZ file (whatever exists)
              sprintf(order2, "rm %s*", order3);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting trace file(s).\n");
                exit(1);
              }
            }
          }

          break;
      }

      // Added: compute averages if multiple slices
      // ( why not directly on test_send ? )
      // ( delay shouldn't be kept ? )
  /*    if (gl.slices > 1)
      {
        w_only_tcp[numbindtcp*(gl.slices)].bandwidth = 0;
        w_only_tcp[numbindtcp*(gl.slices)].delay = 0;
        for (i = 0; i < gl.slices; i ++)
        {
          w_only_tcp[numbindtcp*(gl.slices)].bandwidth += (1/gl.slices)*w_only_tcp[numbindtcp*(gl.slices - 1) + i].bandwidth;
          w_only_tcp[numbindtcp*(gl.slices)].delay += (1/gl.slices)*w_only_tcp[numbindtcp*(gl.slices - 1) + i].delay;
        }
      } */

      // This code is replicated for the case of simultaneous TCP connections
      if (w_only_tcp[numbindtcp*(gl.slices)].delay < min_delay)
        min_delay = w_only_tcp[numbindtcp*(gl.slices)].delay;

      if (w_only_tcp[numbindtcp*(gl.slices)].delay + w_only_tcp[numbindtcp*(gl.slices)].conn_time < min_delay2) // including conn time
        min_delay2 = w_only_tcp[numbindtcp*(gl.slices)].delay + w_only_tcp[numbindtcp*(gl.slices)].conn_time;

      if (w_only_tcp[numbindtcp*(gl.slices)].delay > max_delay)
        max_delay = w_only_tcp[numbindtcp*(gl.slices)].delay;

      if (w_only_tcp[numbindtcp*(gl.slices)].delay + w_only_tcp[numbindtcp*(gl.slices)].conn_time > max_delay2) // including conn time
        max_delay2 = w_only_tcp[numbindtcp*(gl.slices)].delay + w_only_tcp[numbindtcp*(gl.slices)].conn_time;

      if (w_only_tcp[numbindtcp*(gl.slices)].bandwidth > max_bw)
      {
        max_bw = w_only_tcp[numbindtcp*(gl.slices)].bandwidth;
      }

      tot_bw += w_only_tcp[numbindtcp*(gl.slices)].bandwidth;
      numbindtcp++;

    }
    else if (a[i] == interfaces + 1) // MPTCP connection with all interfaces available
    {
      // 2.2. With MPTCP
      //      ( do not care about interfaces )

      if (gl.verbose) fprintf(stdout, "With MPTCP\n");

      switch(gl.ntest)
      {
        case ONEWAY:
          dataport = send_indication(sock_control, ONEWAY);

          if (gl.collect_traces)
          {
            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)    // provisional, only for debugging (with local option) -- Aug 2014
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

              if (gl.debug) printf("[Debug] Tcpdump on interface : %s\n", ifreq[b[k]].ifr_name);
              vpid[k] = launch_tcpdump(ifreq[b[k]].ifr_name, DEV_CLIENT, (int) id_client, 0, k, dataport, order); 
//              vpid[k] = launch_tcpdump(ifreq[b[k]].ifr_name, DEV_CLIENT, 0, k, dataport, order); 
              // vpid[k] = launch_tcpdump(ifreq[b[0]].ifr_name, DEV_CLIENT, 0, k, dataport, order);      // provisional, only for debugging (with local option) -- Aug 2014
            }
          }

//          if ( ( ret = ( gl.nmetric == BW ? test_send_time(&(w_mptcp[0]), gl.ipa, dataport, NOINTERFACE, 1, f_name2) : test_send(&(w_mptcp[0]), gl.ipa, dataport, NOINTERFACE, 1, f_name2) ) ) < 0)
//          if ( ( ret = test_send_cbr(&(w_mptcp[0]), gl.ipa, dataport, NOINTERFACE, 1, f_name2, gl.nmetric) ) < 0)
          if ( ( ret = test_send_cbr(&(w_mptcp[0]), gl.ipa, dataport, NOINTERFACE, 1, f_name2, gl.nmetric, 0) ) < 0)
          {
            fprintf(stdout, "Test 2 (oneway) failed!\n");
            exit(1);
          }

          if (gl.collect_traces)
          {
            sleep(1); // give time to socket closing instructions to complete

            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)  // provisional, only for debugging (with local option) -- Aug 2014  
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

              sprintf(order, "%02d-%010d-client-%1d.pcap", id_client, 0, k);
//              sprintf(order, "%010d-client-%d.pcap", 0, k);

              vret[k] = stop_tcpdump(vpid[k], DEV_CLIENT, order); // if ret!=0, something went wrong in tcpdump or gzip, don't send trace
              
            }

            // Send trace

            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)             // provisional, only for debugging (with local option) -- Aug 2014  
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

              if (vret[k] != 0) // remove tracefiles and continue
              {
//                sprintf(order, "%010d-client-%d.pcap*", 0, k);  
                sprintf(order, "%02d-%010d-client-%1d.pcap*", id_client, 0, k);  

                sprintf(order2, "rm %s", order);  
                if ((ret = system(order2)) < 0 )
                { 
                  fprintf(stdout, "Error when deleting MPTCP trace file.\n");
                  exit(1);
                }
           
                continue;
              }

              // If everything went OK in launch_tcpdump (tcpdump/gzip), then send

              sprintf(order, "%02d-%010d-client-%1d.pcap.gz", id_client, 0, k);  
//              sprintf(order, "%010d-client-%d.pcap.gz", 0, k);  
              sprintf(order2, "client-mp-%1d.pcap.gz", k);      // limited to 32 (size of suffix)

              if (gl.debug) fprintf(stdout, "[Debug] Size of sent string: %d\n", (int) strlen(order2));

              send_filechar(sock_control, TXTYPE_FILE, CLI_FILE, order, order2);

              sprintf(order2, "rm %s", order);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting MPTCP trace file.\n");
                exit(1);
              }
            }
          }

          break;
        case REQRESP:
          dataport = send_indication(sock_control, REQRESP);

          if (gl.collect_traces)
          {
            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)    // provisional, only for debugging (with local option) -- Aug 2014
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

              if (gl.debug) printf("[Debug] Tcpdump on interface : %s\n", ifreq[b[k]].ifr_name);
              vpid[k] = launch_tcpdump(ifreq[b[k]].ifr_name, DEV_CLIENT, (int) id_client, 0, k, dataport, order); 
//              vpid[k] = launch_tcpdump(ifreq[b[k]].ifr_name, DEV_CLIENT, 0, k, dataport, order); 
              // vpid[k] = launch_tcpdump(ifreq[b[0]].ifr_name, DEV_CLIENT, 0, k, dataport, order);      // provisional, only for debugging (with local option) -- Aug 2014
            }
          }

          if ( test_recv(&(w_mptcp[0]), gl.ipa, dataport, NOINTERFACE, 1, f_name2) < 0)
          {
            fprintf(stdout, "Test 2 (reqresp) failed!\n");
            exit(1);
          }

          if (gl.collect_traces)
          {
            sleep(1); // give time to socket closing instructions to complete

            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)  // provisional, only for debugging (with local option) -- Aug 2014  
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

//              sprintf(order, "%010d-client-%d.pcap", 0, k);
              sprintf(order, "%02d-%010d-client-%1d.pcap", id_client, 0, k);

              vret[k] = stop_tcpdump(vpid[k], DEV_CLIENT, order); // if ret!=0, something went wrong in tcpdump or gzip, don't send trace
              
            }

            // Send trace

            for (k = 0; k <= av_ifs; k ++)
            // for (k = 0; k <= 1; k ++)             // provisional, only for debugging (with local option) -- Aug 2014  
            {
              if (! valid_interface(ifreq, b[k], gl.local)) continue;

              if (vret[k] != 0) // remove tracefiles and continue
              {
                sprintf(order, "%02d-%010d-client-%1d.pcap*", id_client, 0, k);  
//                sprintf(order, "%010d-client-%d.pcap*", 0, k);  

                sprintf(order2, "rm %s", order);  
                if ((ret = system(order2)) < 0 )
                { 
                  fprintf(stdout, "Error when deleting MPTCP trace file.\n");
                  exit(1);
                }
           
                continue;
              }

              // If everything went OK in launch_tcpdump (tcpdump/gzip), then send

//              sprintf(order, "%010d-client-%d.pcap.gz", 0, k);  
              sprintf(order, "%02d-%010d-client-%1d.pcap.gz", id_client, 0, k);  
              sprintf(order2, "client-mp-%1d.pcap.gz", k);      // limited to 32 (size of suffix)

              if (gl.debug) fprintf(stdout, "[Debug] Size of sent string: %d\n", (int) strlen(order2));

              send_filechar(sock_control, TXTYPE_FILE, CLI_FILE, order, order2);

              sprintf(order2, "rm %s", order);  
              if ((ret = system(order2)) < 0 )
              { 
                fprintf(stdout, "Error when deleting MPTCP trace file.\n");
                exit(1);
              }
            }
          }

          break;
      }
    }
    else if (a[i] == interfaces + 2) // Note : This should be a 'switch' 
    {
      // This is the place for simultaneous TCP connections (for BW2 metric)
      // (now, replicated code from above)

      if (gl.verbose) fprintf(stdout, "With TCP, multiple connections\n");

      switch(gl.ntest)
      {
        case ONEWAY:

      // We assume ONEWAY transmission for now
      // (Although it should be a 'special' type of ONEWAY (to let server allow simultaneous connections)
      dataport = send_indication(sock_control, ONEWAYSIM);

      doublelocal=0; // set 0 for normal operation

      nlo =0; // only used in mode gl.local
 
      totret = 0;

      if (gl.local || doublelocal)
        for (k = 0; k < av_ifs; k ++)
        {  
          if (gl.local && ! doublelocal && (ifreq[b[k]].ifr_flags & IFF_LOOPBACK))
          {
            nlo = k;
          }

          if (doublelocal && (ifreq[b[k]].ifr_flags & IFF_UP))
          {
            nlo = k;
          }

        }      

//      for (k = 0, j = 0; k < av_ifs + 1; k ++)
      for (k = 0, j = 0; k < (gl.local || doublelocal ? 2 : av_ifs); k ++)
      {
        if (gl.local || doublelocal)
        {
          if ( ! (ifreq[b[k]].ifr_flags & IFF_LOOPBACK) && 0) 
          {
            if (gl.verbose) fprintf(stdout, "Not loopback in mode 'local': leaving...\n");
            continue; // provisional: only use lo interface
          }
        }
        else
        {
          if ((ifreq[b[k]].ifr_flags & IFF_LOOPBACK) || ! (ifreq[b[k]].ifr_flags & IFF_UP)) continue;
        }

// if (strcmp(ifreq[b[k]].ifr_name, "eth0.2")!=0 && strcmp(ifreq[b[k]].ifr_name, "eth0")!=0) continue; //provisional para el probe

        if (b[k] < interfaces + 1 || ( (gl.local || doublelocal) && 1 ) ) // without mptcp         // TO CHECK: should not be av_ifs ? which relationship ?
        {
          /* Create socket */
//          sock_sc[j] = (gl.use6 ? socket(AF_INET6, SOCK_STREAM, 0) : socket(AF_INET, SOCK_STREAM, 0));
//          if ( sock_sc[j] < 0 )
//          {
//            fprintf(stdout, "Unable to create socket.\n");
//            exit(1);
//          }
        
          // printf("Created data socket %d.\n", sock_sc[j]);

          pipe(pipefd[j]); // handle result

          pid_sc[j] = fork();
          switch(pid_sc[j])
          {
            case 0: /* child */
              close(pipefd[j][0]); // close the read-end of the pipe

              if ( ( ret = test_send_cbr(&(w_sim_tcp[j*(gl.slices)]), gl.ipa, dataport, b[k] , 0, f_name1, gl.nmetric, 0) ) < 0 )
              {
                fprintf(stdout, "[%d] Test with regular TCP (over %s) failed! return value = %d\n", getpid(), ( gl.local || doublelocal ? "lo" : ifreq[b[k]].ifr_name ), ret);
                exit(1);
              }

              //printf("[%d] OK, survived test_send_cbr.\n", getpid());
              buff=(u_char*)malloc(sizeof(struct measure));
              if (buff==NULL)
              {
                printf("[%d] Error allocating memory.\n", getpid());
                exit(1);
              }
              memcpy(buff, &(w_sim_tcp[j*gl.slices]), sizeof(struct measure)); // fix ? (for other values of slices)
              printf("[%d] Measured delay %lf, bandwidth %lf.\n", getpid(), w_sim_tcp[j*gl.slices].delay, w_sim_tcp[j*gl.slices].bandwidth); 
              ret = write(pipefd[j][1], buff, sizeof(struct measure));
              if ( ret < sizeof(struct measure)) // not error, but bad
              {
                fprintf(stdout, "[%d] Error, written less bytes than the size of struct measure.\n", getpid());
                exit(1);
              }
              else if (ret < 0) // error
              {
                fprintf(stdout, "[%d] ERROR in writing from child to parent: %s.\n", getpid(), strerror(errno));
                exit(1);
              }

              exit(0);
              break;
            case -1: /* error */
              fprintf(stdout, "ERROR in fork\n");
              break;
            default: /* parent */
              // What if we remove this ?
              close(pipefd[j][1]); // close the write-end of the pipe
              // printf("[%d] Here in the parent.\n", getpid());

          }

          j ++;
        }
      }

//      for (k = 0; k < j; k ++) close(sock_sc[k]);

      // Wait for children
      for (k = 0; k < j; k ++) // j is the number of created sockets (and processes) // replaced "interfaces" with "av_ifs"
      {
        printf("Waiting %d...\n", k);
        // printf("(%d) Before reading from pipe...\n", k);

        buff=(u_char*)malloc(sizeof(struct measure));
        if (buff==NULL)
        {
          printf("[%d] Error allocating memory.\n", getpid());
          exit(1);
        }

// ---------------- make it with select 02.12.2014

        fd_set set;
        struct timeval timeout;

        /* Initialize the file descriptor set. */
        FD_ZERO(&set);
        FD_SET(pipefd[k][0], &set);

        /* Initialize the timeout data structure. */
        timeout.tv_sec = gl.secmin*2; // be liberal in the waiting... (it should terminate at gl.secmin or a bit later)
        timeout.tv_usec = 0;

        /* In the interest of brevity, I'm using the constant FD_SETSIZE, but a more 
         efficient implementation would use the highest fd + 1 instead. In your case
         since you only have a single fd, you can replace FD_SETSIZE with
         pipe_fds[0] + 1 thereby limiting the number of fds the system has to
         iterate over. */
        int ret = select(pipefd[k][0] + 1, &set, NULL, NULL, &timeout);
 
        // a return value of 0 means that the time expired
        // without any acitivity on the file descriptor
        if (ret == 0)
        {
          printf("ERROR ! No reply from the child, abort...\n");
          send_abort(sock_control);
          exit(1); 
        }
        else if (ret < 0)
        {
          // error occurred
          fprintf(stdout, "[%d] ERROR in reading in parent from child: %s.\n", getpid(), strerror(errno));
          exit(1);
        }
        else
        {
          // there was activity on the file descripor
          ret = read(pipefd[k][0], buff, sizeof(struct measure));
        }

// --------------------------------------

// Replaces:
/*
        ret = read(pipefd[k][0], buff, sizeof(struct measure));

        if (ret < sizeof(struct measure))
        {
          ret += read(pipefd[k][0], buff, sizeof(struct measure));
          fprintf(stdout, "[%d] Error, received less bytes than the size of struct measure: in total, %d (should be %d).\n", getpid(), ret, (int) sizeof(struct measure));
          //exit(1);
        }
        else if (ret < 0)
        {
          fprintf(stdout, "[%d] ERROR in reading in parent from child: %s.\n", getpid(), strerror(errno));
          exit(1);
        }
*/
        memcpy(&(w_sim_tcp[k*gl.slices]), buff, sizeof(struct measure)); // fix ? (for other values of slices)
        printf("{ k=%d , Delay: %lf , Bandwidth: %lf }\n", k, w_sim_tcp[k*gl.slices].delay, w_sim_tcp[k*gl.slices].bandwidth); 

        if (waitpid(pid_sc[k], &status, 0) > 0)
        {
          printf("  (Child %d (pid %d) terminated.)\n", k, pid_sc[k]);
          if (status != 0)
          {
            fprintf(stdout, "Not-so-normal termination of child %i (pid %d):", k, pid_sc[k]);
            if (WIFSTOPPED(status)) fprintf(stdout, "  WIFSTOPPED\n");
            if (WIFCONTINUED(status)) fprintf(stdout, "  WIFCONTINUED\n");
            if (WIFSIGNALED(status)) fprintf(stdout, "  WIFSIGNALED (signal %d)\n", WTERMSIG(status));
            if (WIFEXITED(status)) fprintf(stdout, "  WIFEXITED (signal %d)\n", WEXITSTATUS(status));
          }
        }
        else
        {
          fprintf(stdout, "ERROR when waiting for child %d (pid %d): %s.\n", k, pid_sc[k], strerror(errno));
        }

        // Replicated from TCP individual case

        if (w_sim_tcp[k*(gl.slices)].delay < min_delay)
          min_delay = w_sim_tcp[k*(gl.slices)].delay;     // Who cares about the delay when running simultaneous TCP conns?

        if (w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time < min_delay2) // including conn time
          min_delay2 = w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time;
 
        if (w_sim_tcp[k*(gl.slices)].delay > max_delay)
          max_delay = w_sim_tcp[k*(gl.slices)].delay;

        if (w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time > max_delay2) // including conn time
          max_delay2 = w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time;

        if (w_sim_tcp[k*(gl.slices)].bandwidth > max_bw)
        {
          max_bw_sim = w_sim_tcp[k*(gl.slices)].bandwidth;
        }

        tot_bw_sim += w_sim_tcp[k*(gl.slices)].bandwidth;
      }

          break;

        case REQRESP: // (for now, copy from ONEWAY, just before)

      // We assume ONEWAY transmission for now
      // (Although it should be a 'special' type of ONEWAY (to let server allow simultaneous connections)
      dataport = send_indication(sock_control, REQRESPSIM);

      doublelocal=0; // set 0 for normal operation

      nlo = 0; // only used in mode gl.local
 
      totret = 0;

      if (gl.local || doublelocal)
        for (k = 0, j = 0; k < av_ifs; k ++)
        {  
          if (gl.local && ! doublelocal && (ifreq[b[k]].ifr_flags & IFF_LOOPBACK))
          {
            nlo = k;
          }

          if (doublelocal && (ifreq[b[k]].ifr_flags & IFF_UP))
          {
            nlo = k;
          }

        }      

//      for (k = 0, j = 0; k < av_ifs + 1; k ++)
      for (k = 0, j = 0; k < (gl.local || doublelocal ? 2 : av_ifs); k ++)
      {
        if (gl.local || doublelocal)
        {
          if ( ! (ifreq[b[k]].ifr_flags & IFF_LOOPBACK) && 0) 
          {
            if (gl.verbose) fprintf(stdout, "Not loopback in mode 'local': leaving...\n");
            continue; // provisional: only use lo interface
          }
        }
        else
        {
          if ((ifreq[b[k]].ifr_flags & IFF_LOOPBACK) || ! (ifreq[b[k]].ifr_flags & IFF_UP)) continue;
        }

// if (strcmp(ifreq[b[k]].ifr_name, "eth0.2")!=0 && strcmp(ifreq[b[k]].ifr_name, "eth0")!=0) continue; //provisional para el probe

        if (b[k] < interfaces + 1 || ( (gl.local || doublelocal) && 1 ) ) // without mptcp         // TO CHECK: should not be av_ifs ? which relationship ?
        {
          /* Create socket */
//          sock_sc[j] = (gl.use6 ? socket(AF_INET6, SOCK_STREAM, 0) : socket(AF_INET, SOCK_STREAM, 0));
//          if ( sock_sc[j] < 0 )
//          {
//            fprintf(stdout, "Unable to create socket.\n");
//            exit(1);
//          }
        
          // printf("Created data socket %d.\n", sock_sc[j]);

          // dont need it for REQRESPSIM 201114
          pipe(pipefd[j]); // handle result

          pid_sc[j] = fork();
          switch(pid_sc[j])
          {
            case 0: /* child */
              close(pipefd[j][0]); // close the read-end of the pipe

/* More changes in reqresp: pipes. One thing is writing, another is reading, (?) (20/11/2014) */

//            THIS changes in reqresp !
//              if ( ( ret = test_send_cbr(&(w_sim_tcp[j*(gl.slices)]), gl.ipa, dataport, b[k] , 0, f_name1, gl.nmetric, 0) ) < 0 )
              if ( ( ret = test_recv_unif(&(w_sim_tcp[j*(gl.slices)]), gl.ipa, dataport, b[k], 0, f_name1, gl.nmetric) ) < 0 )
              {
                fprintf(stdout, "[%d] Test with regular TCP (over %s) failed! return value = %d\n", getpid(), ( gl.local || doublelocal ? "lo" : ifreq[b[k]].ifr_name ), ret);
                exit(1);
              }

              //printf("[%d] OK, survived test_send_cbr.\n", getpid());
              buff=(u_char*)malloc(sizeof(struct measure));
              if (buff==NULL)
              {
                printf("[%d] Error allocating memory.\n", getpid());
                exit(1);
              }
              memcpy(buff, &(w_sim_tcp[j*gl.slices]), sizeof(struct measure)); // fix ? (for other values of slices)
              printf("[%d] Measured delay %lf, bandwidth %lf.\n", getpid(), w_sim_tcp[j*gl.slices].delay, w_sim_tcp[j*gl.slices].bandwidth); 
              ret = write(pipefd[j][1], buff, sizeof(struct measure));
              if ( ret < sizeof(struct measure)) // not error, but bad
              {
                fprintf(stdout, "[%d] Error, written less bytes than the size of struct measure.\n", getpid());
                exit(1);
              }
              else if (ret < 0) // error
              {
                fprintf(stdout, "[%d] ERROR in writing from child to parent: %s.\n", getpid(), strerror(errno));
                exit(1);
              }

              exit(0);
              break;
            case -1: /* error */
              fprintf(stdout, "ERROR in fork\n");
              break;
            default: /* parent */
              // What if we remove this ?
              close(pipefd[j][1]); // close the write-end of the pipe
              // printf("[%d] Here in the parent.\n", getpid());
          }

          j ++;
        }
      }

//      for (k = 0; k < j; k ++) close(sock_sc[k]);

      // Wait for children
      for (k = 0; k < j; k ++) // j is the number of created sockets (and processes) // replaced "interfaces" with "av_ifs"
      {
        printf("Waiting %d...\n", k);
        // printf("(%d) Before reading from pipe...\n", k);

              buff=(u_char*)malloc(sizeof(struct measure));
              if (buff==NULL)
              {
                printf("[%d] Error allocating memory.\n", getpid());
                exit(1);
              }

        ret = read(pipefd[k][0], buff, sizeof(struct measure));

        if (ret < sizeof(struct measure))
        {
          ret += read(pipefd[k][0], buff, sizeof(struct measure));
          fprintf(stdout, "[%d] Error, received less bytes than the size of struct measure: in total, %d (should be %d).\n", getpid(), ret, (int) sizeof(struct measure));
          //exit(1);
        }
        else if (ret < 0)
        {
          fprintf(stdout, "[%d] ERROR in reading in parent from child: %s.\n", getpid(), strerror(errno));
          exit(1);
        }

        memcpy(&(w_sim_tcp[k*gl.slices]), buff, sizeof(struct measure)); // fix ? (for other values of slices)
        printf("{ k=%d , Delay: %lf , Bandwidth: %lf }\n", k, w_sim_tcp[k*gl.slices].delay, w_sim_tcp[k*gl.slices].bandwidth); 

        if (waitpid(pid_sc[k], &status, 0) > 0)
        {
          printf("  (Child %d (pid %d) terminated.)\n", k, pid_sc[k]);
          if (status != 0)
          {
            fprintf(stdout, "Not-so-normal termination of child %i (pid %d):", k, pid_sc[k]);
            if (WIFSTOPPED(status)) fprintf(stdout, "  WIFSTOPPED\n");
            if (WIFCONTINUED(status)) fprintf(stdout, "  WIFCONTINUED\n");
            if (WIFSIGNALED(status)) fprintf(stdout, "  WIFSIGNALED (signal %d)\n", WTERMSIG(status));
            if (WIFEXITED(status)) fprintf(stdout, "  WIFEXITED (signal %d)\n", WEXITSTATUS(status));
          }
        }
        else
        {
          fprintf(stdout, "ERROR when waiting for child %d (pid %d): %s.\n", k, pid_sc[k], strerror(errno));
        }

        // Replicated from TCP individual case

        if (w_sim_tcp[k*(gl.slices)].delay < min_delay)
          min_delay = w_sim_tcp[k*(gl.slices)].delay;     // Who cares about the delay when running simultaneous TCP conns?

        if (w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time < min_delay2) // including conn time
          min_delay2 = w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time;
 
        if (w_sim_tcp[k*(gl.slices)].delay > max_delay)
          max_delay = w_sim_tcp[k*(gl.slices)].delay;

        if (w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time > max_delay2) // including conn time
          max_delay2 = w_sim_tcp[k*(gl.slices)].delay + w_sim_tcp[k*(gl.slices)].conn_time;

        if (w_sim_tcp[k*(gl.slices)].bandwidth > max_bw)
        {
          max_bw_sim = w_sim_tcp[k*(gl.slices)].bandwidth;
        }

        tot_bw_sim += w_sim_tcp[k*(gl.slices)].bandwidth;
      }



          break;
      }

    }
  }

  // 2b. Sending string with slices results

  if (gl.verbose) fprintf(stdout, "\nRESULTS COMPUTATION AND TRANSMISSION\n");

  if (gl.slices > 1)
  { 
    sprintf(tsaf, "Slices results ; format: n_slice [delay-tcp bytes-tcp (foreach av.if.)] [delay-mptcp bytes-mptcp]\n--------------\n");

    for (i=0; i<gl.slices; i++)
    {
      sprintf(addstr, "%d ", i);
      concatenate_string(tsaf, addstr);
 
      for (j=0; j<av_ifs; j++)
      {
        sprintf(addstr, "%lf %ld ", w_only_tcp[j*(gl.slices)+i].delay, w_only_tcp[j*(gl.slices)+i].bytes);
        concatenate_string(tsaf, addstr);
      }
      sprintf(addstr, "%lf %ld", w_mptcp[i].delay, w_mptcp[i].bytes);
      concatenate_string(tsaf, addstr);

      if (i==0) 
      {
        sprintf(addstr, " # totals");
        concatenate_string(tsaf, addstr);    
      }

      sprintf(addstr, "\n");
      concatenate_string(tsaf, addstr);    
    }

    send_filechar(sock_control, TXTYPE_STRING, CLI_FILE, tsaf, "slices.txt");
  }

  // 3. Computing BW and Delay Aggregation Benefits (see D2.1 for specification)

  if (gl.nmetric == BW) {
    bw_benefit = compute_bw_benefit( w_mptcp[0].bandwidth, max_bw, tot_bw );
  } else if (gl.nmetric == BW2) {
    bw_benefit = compute_bw_benefit( w_mptcp[0].bandwidth, max_bw_sim, tot_bw_sim );
  } else if (gl.nmetric == BWALL) {
    bw_benefit = compute_bw_benefit( w_mptcp[0].bandwidth, max_bw, tot_bw_sim );
    tdi = compute_tdindex(tot_bw, tot_bw_sim, max_bw);
  } else if (gl.nmetric == DELAY) {
    delay_benefit = compute_delay_benefit(w_mptcp[0].delay, max_delay, min_delay, max_delay/min_delay);
    delay2_benefit = compute_delay_benefit(w_mptcp[0].delay + w_mptcp[0].conn_time, max_delay2, min_delay2, max_delay2/min_delay2); // including conn times
  }

  /* Check values */
  if ( (abs(delay_benefit) > 1 || abs(bw_benefit) > 1 ) && gl.nmetric == DELAY )
  {
    fprintf(stdout, "  Warning ! Incorrect benefit value, outside interval [-1, 1].\n");
    if (gl.verbose || gl.debug) printf("    For debug: w_mptcp[0].bandwidth=%f, max_bw=%f, tot_bw=%f\n", w_mptcp[0].bandwidth, max_bw, tot_bw);   
  }

  // 3. Final report : bw benefit and delay benefit
  if (gl.nmetric == BW || gl.nmetric == BW2 || gl.nmetric == BWALL)
  {
    if (gl.verbose) fprintf(stdout, "MPTCP BW Aggregation Benefit    : %lf\n", bw_benefit);
    if (gl.nmetric == BWALL) if (gl.verbose) fprintf(stdout, "Transport Disjointivity Index   : %lf\n", tdi);
    // if (gl.nmetric == BWALL) if (gl.verbose) fprintf(stdout, "tot_bw_sim   : %lf\n", tot_bw_sim); //provisional, remove when possible
    
  }
  else if (gl.nmetric == DELAY)
  {
    if (gl.verbose) fprintf(stdout, "MPTCP Delay Aggregation Benefit : %lf\n", delay_benefit);
    if (gl.verbose) fprintf(stdout, "MPTCP Delay Aggregation Benefit : %lf (including connection time)\n", delay2_benefit);
  }

  // 4. Create "output line" to the server

  // Example of output line:
  // 120        1mbit    142   0.95528687  0.97096171 1.86261986  0.933393      tc qdisc add dev eth0.2 root handle 1: tbf rate 1mbit latency 100ms burst 1540

/*
  Try to rationalize this...

  switch(gl.nmetrics)
  {
    case BW:      // to use: linheader
      sprintf(linheader, "%20s %03d %15s %7s %2d %3d ", gl.tag, id_client, "bw", (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
      break;
    case DELAY:   // to use: linheader linheader2
      sprintf(linheader, "%20s %03d %15s %7s %2d %3d ", gl.tag, id_client, "delay", (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
      sprintf(lin2header, "%20s %03d %15s %7s %2d %3d ", gl.tag, id_client, "delay2", (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
      break;
  }
*/

  // sprintf(outpserver, "%10s %03d %5s %7s %d %d ", gl.tag, id_client, (gl.nmetric == BW ? "bw" : "delay"), (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
  sprintf(outpserver, "%20s %03d %15s %7s %2d %3d ", gl.tag, id_client, avmetrics[gl.nmetric - 1], (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
  if (gl.nmetric == DELAY) sprintf(initial, "%20s %03d %15s %7s %2d %3d ", gl.tag, id_client, "delay2", (gl.ntest == ONEWAY ? "oneway" : "reqresp"), av_ifs, gl.slices);
  //strcpy(initial, outpserver);
  sprintf(outpserver2, "%20s %03d ", gl.tag, id_client);

  for (i = 0; i < (gl.nmetric == BWALL ? av_ifs + 2 : av_ifs + 1); i++)
//  for (i = 0; i < av_ifs + 1; i++)
  {
    sprintf(addstr, "%d", a[i]);
    concatenate_string(outpserver, addstr);
    if (gl.nmetric == DELAY) concatenate_string(initial, addstr);

    // PENDING : Need to include CBRBYTES and CBRTIME metrics !!
    if (gl.nmetric == BW)
    {
      if (i < av_ifs) // only tcp
      {
        sprintf(addstr2, " %lf ", w_only_tcp[i*(gl.slices)].bandwidth); 
      } 
      else if (i==av_ifs)           // mptcp
      {
        sprintf(addstr2, " %lf ", w_mptcp[0].bandwidth);       
      }
    }
    else if (gl.nmetric == BW2)
    {
      if (i < av_ifs) // only tcp
      {
        sprintf(addstr2, " %lf ", w_sim_tcp[i*(gl.slices)].bandwidth); 
      } 
      else if (i==av_ifs)           // mptcp
      {
        sprintf(addstr2, " %lf ", w_mptcp[0].bandwidth);       
      }
    }
    else if (gl.nmetric == BWALL)
    {
      if (i < av_ifs) // only tcp
      {
        sprintf(addstr2, " %lf ", w_only_tcp[i*(gl.slices)].bandwidth); 

        if (w_only_tcp[i*(gl.slices)].bandwidth == 0) { send_abort(sock_control); printf("Inconsistent value of bandwidth, something went wrong. i=%d, gl.slices=%d, bw %lf, delay %lf. Aborting...\n", i, gl.slices, w_only_tcp[i*(gl.slices)].bandwidth, w_only_tcp[i*(gl.slices)].delay); exit(1); }
      } 
      else if (i==av_ifs)
      {
        sprintf(addstr2, " %lf %lf ", tot_bw_sim, w_mptcp[0].bandwidth);       
      }

/*      else if (i == av_ifs) // total tcp
      {
        sprintf(addstr2, " %lf ", tot_bw_sim);       
      }
      else if (i == av_ifs + 1) // mptcp
      {
        sprintf(addstr2, " %lf ", w_mptcp[0].bandwidth);       
      }
*/
    }
    else if (gl.nmetric == DELAY)// DELAY
    {
      if (i < av_ifs) // only tcp
      {
        sprintf(addstr2, " %lf ", w_only_tcp[i*(gl.slices)].delay); 
        sprintf(addstr2b, " %lf ", w_only_tcp[i*(gl.slices)].delay + w_only_tcp[i*(gl.slices)].conn_time); 
      } 
      else if (i==av_ifs)           // mptcp
      {
        sprintf(addstr2, " %lf ", w_mptcp[0].delay);       
        sprintf(addstr2b, " %lf ", w_mptcp[0].delay + w_mptcp[0].conn_time);       
      }
    }
    
    if (i < av_ifs + 1)
    {
      concatenate_string(addstr3, addstr2);
      concatenate_string(addstr3b, addstr2b);
    }

    // This is another chunk of text
    if (i < av_ifs) // only tcp
    {
      sprintf(addstr2, " %lf ", w_only_tcp[i*(gl.slices)].conn_time); 
    } 
    else if (i==av_ifs)           // mptcp
    {
      sprintf(addstr2, " %lf", w_mptcp[0].conn_time);
//      sprintf(addstr2, " %lf \n", w_mptcp[0].conn_time);
    }
    concatenate_string(outpserver2, addstr2);
    //if (gl.nmetric == DELAY) concatenate_string(initial, addstr2);

  }
  concatenate_string(outpserver, addstr3);

  if (gl.nmetric == BW || gl.nmetric == BW2 || gl.nmetric == BWALL) {
    sprintf(addstr, " %3d ", gl.secmin); 
  } else if (gl.nmetric == DELAY) {
    concatenate_string(initial, addstr3b);

    sprintf(addstr, " %9ld ", (long) gl.to_transmit);
    concatenate_string(initial, addstr);  
  }

  concatenate_string(outpserver, addstr);

  if (gl.nmetric == BW || gl.nmetric == BW2 || gl.nmetric == BWALL) 
  {
    sprintf(addstr, "%10f", bw_benefit); 
//    sprintf(addstr, "%10f\n", bw_benefit); 
  }
  else if (gl.nmetric == DELAY)
  {
    sprintf(addstr, "%10f", delay_benefit);
//    sprintf(addstr, "%10f\n", delay_benefit);
    sprintf(addstr2, "%10f", delay2_benefit);
//    sprintf(addstr2, "%10f\n", delay2_benefit);
    concatenate_string(initial, addstr2);
//    concatenate_string(addstr, initial);
  }
 //sprintf(addstr, "%10f %10f\n", bw_benefit, delay_benefit);
  concatenate_string(outpserver, addstr);

  // Note: \n included in string

  if (gl.debug) printf("[Debug] %s\n[Debug] %s\n", outpserver, outpserver2);

  if (av_ifs > 1) { // otherwise you are not providing any useful information
//    printf("[Debug] SENDING outpserver [%s]\n", outpserver);
    send_filechar(sock_control, TXTYPE_STRING, CLI_LOGLINE, outpserver, "log2.txt");
    if (gl.nmetric == DELAY) {
//      printf("[Debug] SENDING initial [%s]\n", initial);
      send_filechar(sock_control, TXTYPE_STRING, CLI_LOGLINE, initial, "log2.txt");    // repeated, for delay2 (including conn times)
    }
//    printf("[Debug] SENDING outpserver2 [%s]\n", outpserver2);

    send_filechar(sock_control, TXTYPE_STRING, CLI_LOGLINE, outpserver2, "log_tcet.txt");
  }

  // 5. ...send

  send_result(sock_control, bw_benefit, delay_benefit);

  /* Restoring original values */

  memset(order, 0, BUFFSIZE); 

  // Pending: check permission (well, we are forcing root privileges)
  sprintf(order, "sysctl -w net.mptcp.mptcp_enabled=%d > /dev/null", old_mp_enabled);
  if ((i = system(order)) < 0 )
  { 
    fprintf(stdout, "Error when restoring value for mptcp_enabled.\n");
    exit(1);
  }

  // Close control socket
  if (close(sock_control) != 0)
  {
    printf("ERROR, control socket could not be successfully closed.\n");
  } 
  
  return 0;
}

void send_result(int sock_control, double bw_benefit, double delay_benefit)
{
  struct control_header *testdef;
  struct res_benefit *report;
  u_char* databuff;
  int ret;

  int sock;

  /* Define control packet contents (in struct) */

  testdef = malloc(sizeof(struct control_header));
  if (testdef == NULL) 
  {
    fprintf(stdout, "Error allocating memory for testdef.\n");
    exit(1);
  }

  report = malloc(sizeof(struct res_benefit));
  if (report == NULL) 
  {
    fprintf(stdout, "Error allocating memory for report.\n");
    exit(1);
  }

  testdef->device = DEV_CLIENT;      // 0x01 DEV_CLIENT, 0x02 DEV_SERVER
  testdef->type = htons(CLI_REPORT); // 0x0102 finalreport
  testdef->id_client = id_client;    // char

  sprintf(report->bw_value, "%-1.4f", bw_benefit);
  sprintf(report->delay_value, "%-1.4f", delay_benefit);
//  sprintf(report->conntime_value, "%6f", conn_time*1000000); // Transmitted value in usec

  /* Serialization */

  /* Note : Careful with the use of sizeof(struct) for serialization or socket transmission. 
     Due to data alignment, it can be bigger than the size of the data actually allocated. 
     Padding currently disabled with #pragma pack(1), portable?
  */

  databuff = (u_char*)malloc(sizeof(struct control_header) + sizeof(struct res_benefit));
  if (databuff == NULL) 
  {
    fprintf(stdout, "Error allocating memory for databuff.\n");
    exit(1);
  }

  memcpy(databuff, testdef, sizeof(struct control_header));
  memcpy(databuff + sizeof(struct control_header), report, sizeof(struct res_benefit));

  /* Only for debugging */

  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent header: device %02x, type %04x, id_client %02x ]\n", time_to_seconds( &t0, &now ), testdef->device, ntohs(testdef->type), testdef->id_client); }
  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent report: bw benefit %s, delay benefit %s ]\n", time_to_seconds( &t0, &now ), report->bw_value, report->delay_value); }

  free(testdef);
  free(report);

  /* Send report */

  if ((ret = send(sock_control, databuff, sizeof(struct control_header) + sizeof(struct res_benefit), 0)) < 0)
  {
    fprintf(stdout, "Could not send report.\n");
    exit(1);
  }

  // (Implicit in the packet log)
  // if (gl.verbose) fprintf(stdout, "Report sent successfully (ret=%d).\n", ret);
  // if (gl.verbose) show_hexa(databuff, sizeof(struct control_header)+sizeof(struct res_benefit));
}

void send_abort(int sock_control)
{
  struct control_header *testdef;
  u_char* databuff;
  int ret;

  int sock;

  /* Define control packet contents (in struct) */

  testdef = malloc(sizeof(struct control_header));
  if (testdef == NULL) 
  {
    fprintf(stdout, "Error allocating memory for testdef.\n");
    exit(1);
  }

  testdef->device = DEV_CLIENT;      // 0x01 DEV_CLIENT, 0x02 DEV_SERVER
  testdef->type = htons(CLI_ABORT);  // 0x01FF abort
  testdef->id_client = id_client;    // char

  /* Serialization */

  /* Note : Careful with the use of sizeof(struct) for serialization or socket transmission. 
     Due to data alignment, it can be bigger than the size of the data actually allocated. 
     Padding currently disabled with #pragma pack(1), portable?
  */

  databuff = (u_char*)malloc(sizeof(struct control_header));
  if (databuff == NULL) 
  {
    fprintf(stdout, "Error allocating memory for databuff.\n");
    exit(1);
  }

  memcpy(databuff, testdef, sizeof(struct control_header));

  /* Only for debugging */

  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent header: device %02x, type %04x, id_client %02x ]\n", time_to_seconds( &t0, &now ), testdef->device, ntohs(testdef->type), testdef->id_client); }

  free(testdef);

  /* Send report */

  if ((ret = send(sock_control, databuff, sizeof(struct control_header) + sizeof(struct res_benefit), 0)) < 0)
  {
    fprintf(stdout, "Could not send abort.\n");
    exit(1);
  }
}


//int establish(int srcport, char * dstipa, int dstport, int nbr_if, double* conn_time)
int establish(int srcport, char * dstipa, int dstport, int nbr_if, double* conn_time, int sock_ex)
{
  int i, j, sock;
  int yes=1;

  struct timeval t0, t1;

  struct ifreq ifr;
  
  if (sock_ex == 0) // Then, create the new socket; otherwise, use socket_ex
  {
    /* Create socket */
    sock = (gl.use6 ? socket(AF_INET6, SOCK_STREAM, 0) : socket(AF_INET, SOCK_STREAM, 0));
    if ( sock < 0 )
    {
      fprintf(stdout, "Unable to create socket.\n");
      exit(1);
    }
  }
  else
  {
    sock = sock_ex;
  }

  //if (gl.verbose) fprintf(stdout, "Created socket.\n");

//=====================================

  if (srcport != 0) // I think all this is never used (since we don't bind to specific srcports in client)
  {

    if (!gl.use6)
    {
      /* Note: sizeof(conn) -conn is a pointer- or sizeof(struct sockaddr_in) (or, equivalently, sizeof(*conn)) ? */ 

      struct sockaddr_in * conn; // = (struct sockaddr_in *) parsed_conn;
      conn = (struct sockaddr_in *) malloc(sizeof(struct sockaddr_in));
      //printf("sizeof conn %ld , sizeof *conn %ld\n", sizeof(conn), sizeof(*conn));

      memset((struct sockaddr_in*) conn, 0, sizeof(* conn));              // Clear struct
      conn->sin_family = AF_INET;                 // IPv4 address family
      conn->sin_addr.s_addr = htonl(INADDR_ANY);  // Accepting incoming packets to ANY IPA (from the host)
      conn->sin_port = htons(srcport);            // port (not global)

/*
      // Bind 
      if ( bind(sock, (struct sockaddr*) conn, sizeof(*conn)) < 0 ) 
      {
        printf("Unable to bind socket to src port %d.\n", srcport);
        exit(1);
      }
*/

      // Bind 
      i=-1;
      j=0;
      do  
      {
        i = bind(sock, (struct sockaddr*) conn, sizeof(*conn));
        j ++;
        if (gl.verbose) printf(".");
        sleep(2);
      } while (i < 0 && j < 10);

      if (i < 0)
      {
        printf("Unable to bind socket to port %d.\n", srcport);
        exit(1);
      }

    }
    else if (gl.use6) // forget about this for now
    {  
      struct sockaddr_in6 * conn; // = (struct sockaddr_in6 *) parsed_conn;
      memset(conn, 0, sizeof(* conn));              // Clear struct
      conn->sin6_family = AF_INET6;                // IPv6 address family
      conn->sin6_addr = in6addr_any;               // Incoming IPA: any
        // previous: htons(INADDR_ANY)
      conn->sin6_port = htons(srcport);               // port

/* Old:
      // Bind 
      if ( bind(sock, (struct sockaddr*) conn, sizeof(*conn)) < 0 ) 
      {
        printf("Unable to bind socket to port %d.\n", srcport);
        exit(1);
      }
*/

      // Bind 
      i=-1;
      j=0;
      do  
      {
        i = bind(sock, (struct sockaddr*) conn, sizeof(*conn));
        j ++;
        if (gl.verbose) printf(".");
        sleep(1);
      } while (i < 0 || j < 3);

      if (i < 0)
      {
        printf("Unable to bind socket to port %d.\n", srcport);
        exit(1);
      }

    }
  }

//===================================

  /* Interface binding */

  if (nbr_if != NOINTERFACE)
  {
    memset(&ifr, 0, sizeof(ifr));                 // Clear struct
    if (gl.local)
      snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "lo");
    else
      strcpy(ifr.ifr_name, ifreq[nbr_if].ifr_name);
//      snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), ifreq[nbr_if].ifr_name);

    char * used_ipa;
    used_ipa = (char *) malloc(INET_ADDRSTRLEN * sizeof(char));
    inet_ntop(AF_INET, &(( (struct sockaddr_in *) &(ifreq[nbr_if]).ifr_addr )->sin_addr), used_ipa, INET_ADDRSTRLEN ); // and for IPv6 ... ?

    if (gl.verbose) fprintf(stdout, "[%d] Using %s (%s)\n", getpid(), ifreq[nbr_if].ifr_name, used_ipa );
//    fprintf(stdout, "Using %s (%s)\n", ifreq[nbr_if].ifr_name, inet_ntoa( ifreq[nbr_if].ifr_addr.sin_addr) );
//    fprintf(stdout, "Using %s (%s)\n", ifreq[nbr_if].ifr_name, inet_ntoa(( (struct sockaddr_in *) &(ifreq[nbr_if]).ifr_addr )->sin_addr) );
    // ...and in IPv6 ?

/*
    // Setsockopt to reuse
    if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1 )
    {
      printf("Unable to configure reusing option in socket.\n");
    }
*/

    if (gl.local)
    {
      i=setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, (void *)&ifr, sizeof(ifr));
//      j=i;
    }
    else
    {
      i=setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, (void *)&ifreq[nbr_if], sizeof(ifreq[nbr_if]));
   //   j=setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&ifreq[nbr_if], sizeof(ifreq[nbr_if]));
    }

//    if (i < 0 || j < 0) 
    if (i < 0) 
    {
      fprintf(stdout, "Unable to bind socket to interface %s (make sure you have root privileges).\n", ifr.ifr_name);
      exit(1);
    }
    // else
    // {
    //   fprintf(stdout, "Successfully binded socket to interface %s.\n", ifr.ifr_name);
    // }

  }

  /* Establish connection */

  if (!gl.use6)  // v4
  {
    struct sockaddr_in conn;

    memset(&conn, 0, sizeof(conn));               // Clear struct
    conn.sin_family = AF_INET;                    // IPv4 address family
    if ( inet_pton(AF_INET, dstipa, &(conn.sin_addr), INET_ADDRSTRLEN) <= 0)  // this changes in v6
    {
      fprintf(stdout, "Error in inet_pton (IPA %s, inet_pton returns %d).\n", dstipa, inet_pton(AF_INET, dstipa, &(conn.sin_addr), INET_ADDRSTRLEN));
      exit(1);
    }

    conn.sin_port = htons(dstport);               // port: 1234
 
    get_now(&t0, 0);    
    i = connect(sock, (struct sockaddr*) &conn, sizeof(conn));
    get_now(&t1, 0);    
  }
  else if (gl.use6) // v6
  {
    struct sockaddr_in6 conn;

    memset(&conn, 0, sizeof(conn));                 // Clear struct
    conn.sin6_family = AF_INET6;                    // IPv6 address family
    conn.sin6_port = htons(dstport);               // port: 1234
    if ( inet_pton(AF_INET6, dstipa, &(conn.sin6_addr), INET6_ADDRSTRLEN) <= 0)
    {
      fprintf(stdout, "Error in inet_pton (IPA %s, inet_pton returns %d).\n", dstipa, inet_pton(AF_INET6, dstipa, &(conn.sin6_addr), INET6_ADDRSTRLEN));
      exit(1);
    }

    get_now(&t0, 0);    
    i = connect(sock, (struct sockaddr*) &conn, sizeof(conn));
    get_now(&t1, 0);    
  }

  if ( i < 0 )
  {
    fprintf(stdout, "Unable to establish connection to %s, port %d, connect=%d.\n",dstipa,dstport,i);
    exit(1);
  }

  if (conn_time != NULL) *conn_time = time_to_seconds(&t0, &t1);

  if (gl.verbose) fprintf(stdout, "[%d] Established connection to %s, port %d (connection establishment time: %.1lf usec).\n", getpid(), dstipa, dstport, time_to_seconds(&t0, &t1)*1000000);

  return sock;
}

int send_filechar(int sock_control, int txtype, int stortype, char* string, char* suffix)
{
  struct control_header *head;
  struct file_tx *filechar;

  struct nbr_port * ind_port;
  struct nbr_bytes * bytes_ind;
  u_char* databuff;
  char buffer[BUFFSIZE];
  int ret;
  int i;

  FILE * f;

  /* Define control packet contents (in struct) */

  head = (struct control_header*)malloc(sizeof(struct control_header));
  if (head == NULL) 
  {
    fprintf(stdout, "Error allocating memory for header.\n");
    exit(1);
  }

  filechar = (struct file_tx *)malloc(sizeof(struct file_tx));
  if (filechar == NULL) 
  {
    fprintf(stdout, "Error allocating memory for file_tx.\n");
    exit(1);
  }

  head->device = DEV_CLIENT;      // 0x01 client, 0x02 server
//  head->type = htons(CLI_FILE);   // 0x0104 cli_file
  head->type = htons(stortype);   // 0x0104 cli_file ; 0x0105 cli_logline
  head->id_client = id_client;    // char

  if (txtype == TXTYPE_FILE)
  {

    f = fopen(string, "r");
    if (f == NULL)
    {
      fprintf(stdout, "ERROR, file does not exist.\n");
      return -1;
    }
    fseek(f, 0, SEEK_END);
 
    filechar->bytes = htonl(ftell(f));
    strcpy(filechar->suffix, suffix); 

    fclose(f);

  }
  else if (txtype == TXTYPE_STRING)
  {
//concatenate_string(string, "\n");
//    for (i=0; string[i]!='\0'; i++);
    filechar->bytes = htonl(strlen(string)); // htonl(i-1);
    strcpy(filechar->suffix, suffix);
//    filechar->type=0x1; // text (to be specified)
  } 

  /* Serialization */

  /* Note : Careful with the use of sizeof(struct) for serialization or socket transmission. 
     Due to data alignment, it can be bigger than the size of the data actually allocated. 
     Padding currently disabled with #pragma pack(1), portable?
  */

  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent header: device %02x, type %04x, id_client %02x ]\n", time_to_seconds( &t0, &now ), head->device, ntohs(head->type), head->id_client); }
  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent file characteristics: bytes %d, suffix %s ]\n", time_to_seconds( &t0, &now ), ntohl(filechar->bytes), filechar->suffix); }

  //ntohl(bytes_ind->bytes_value), ntohl(bytes_ind->slices_value)); }

  databuff = (u_char*)malloc(sizeof(struct control_header) + sizeof(struct file_tx));
  if (databuff == NULL) 
  {
    fprintf(stdout, "Error allocating memory for databuff.\n");
    exit(1);
  }

  //printf("Size of malloc'd databuff : %ld = %ld + %ld (%ld)\n", sizeof(databuff), sizeof(struct control_header), sizeof(struct nbr_bytes), sizeof(u_char));

  memcpy(databuff, head, sizeof(struct control_header));
  memcpy(databuff + sizeof(struct control_header), filechar, sizeof(struct file_tx));
  free(head);

  // printf("Size of malloc'd databuff : %ld = %ld + %ld (%ld)\n", sizeof(databuff), sizeof(struct control_header), sizeof(struct nbr_bytes), sizeof(u_char));

  // if (gl.verbose) show_hexa(databuff, sizeof(struct control_header) + sizeof(struct file_tx));

  if ((ret = send(sock_control, databuff, sizeof(struct control_header) + sizeof(struct file_tx), 0)) < 0)
  {
    fprintf(stdout, "Could not send control packet, ret=%d.\n", ret);
    exit(1);
  }

  if (txtype == TXTYPE_FILE)
  {
    f = fopen(string, "r"); // Second time file is opened
    if (f == NULL)
    {
      fprintf(stdout, "Could not open (for 2nd time) the file to be transmitted.\n");
      exit(1);    
    }
  }

  int total_sent=0;

  memset(buffer, 0, BUFFSIZE);

  int f_block_sz, totsent;

  if (txtype == TXTYPE_FILE)
  {
    while( (f_block_sz = fread(buffer, sizeof(char), BUFFSIZE, f)) > 0 )
    {

      // printf("f_block_sz: %d\n", f_block_sz);

      if( ( ret = send(sock_control, buffer, f_block_sz, 0) ) < 0)
      {
        fprintf(stdout, "ERROR: Failed to send file over control channel.\n");
        break;
      }
      //printf("ret: %d\n", ret);

      memset(buffer, 0, BUFFSIZE);
    }

    //printf("f_block_sz: %d\n", f_block_sz);
  }
  else if (txtype == TXTYPE_STRING)
  {
    totsent = 0;
    /* Actually, by its current definition, string is limited to BUFFSIZE, so only
       one send will be performed, ALWAYS. */
    do
    {
      if (ntohl(filechar->bytes) - totsent < BUFFSIZE) f_block_sz = ntohl(filechar->bytes) - totsent; else f_block_sz = BUFFSIZE;
      if( ( ret = send(sock_control, string, f_block_sz, 0) ) < 0)
      {
        fprintf(stdout, "ERROR: Failed to send file over control channel.\n");
        break;
      }

      totsent += ret;
      string += ret;

    } while (totsent < ntohl(filechar->bytes));
  }

  free(databuff);

  // if (gl.verbose) fprintf(stdout, "Control packet sent successfully (ret=%d).\n", ret);

  return 0;
}

int send_indication(int sock_control, int experiment)
{
  struct control_header *testdef;
  struct nbr_port * ind_port;
  struct nbr_bytes *bytes_ind;
  u_char* databuff;
  int ret;

  /* Define control packet contents (in struct) */

  testdef = (struct control_header*)malloc(sizeof(struct control_header));
  if (testdef == NULL) 
  {
    fprintf(stdout, "Error allocating memory for testdef.\n");
    exit(1);
  }

  bytes_ind = malloc(sizeof(struct nbr_bytes));
  if (bytes_ind == NULL) 
  {
    fprintf(stdout, "Error allocating memory for bytes_ind.\n");
    exit(1);
  }

  testdef->device = DEV_CLIENT;      // 0x01 client, 0x02 server
  switch(experiment)
  {
    case ONEWAY:
      testdef->type = htons(CLI_ONEWAY_INSTRUCTION); // 0x0101 oneway, 0x0201 reqresp
      break;
    case ONEWAYSIM: // new 31/10/2014
      testdef->type = htons(CLI_ONEWAYSIM_INSTRUCTION); // 0x0106 onewaysim
      break;
    case REQRESPSIM: // new 20/11/2014
      testdef->type = htons(CLI_REQRESPSIM_INSTRUCTION); // 0x0107 onewaysim
      break;
    case REQRESP:
      testdef->type = htons(CLI_REQRESP_INSTRUCTION); // 0x0101 oneway, 0x0201 reqresp
      break;
    default: // Unknown experiment !
      printf("ERROR ! Unknown experiment, no indication will be sent.\n");
      return 255;
  }
//  testdef->type = htons((experiment == ONEWAY ? CLI_ONEWAY_INSTRUCTION : CLI_REQRESP_INSTRUCTION)); // 0x0101 oneway, 0x0201 reqresp
  testdef->id_client = id_client;    // char

  bytes_ind->metric = htons(gl.nmetric);              // new 05.09

  switch(gl.nmetric)
  {
    case BWALL: // as in BW 
    case BW2: // as in BW 
    case BW:
      bytes_ind->bytes_value = 0;
      bytes_ind->slices_value = htonl((u_int32_t) gl.avifs);   // if bw, no slices are allowed; thus use it for number of ofs (useful for bwall, in onewaysim and reqrespsim exp.)
      // bytes_ind->slices_value = htonl((u_int32_t) gl.slices);
      bytes_ind->cbr_value = 0;
      bytes_ind->timettx_value = htonl(gl.secmin);
      break;
    case DELAY:
      bytes_ind->bytes_value = htonl(gl.to_transmit);
      bytes_ind->slices_value = htonl((u_int32_t) gl.slices);
      bytes_ind->cbr_value = 0;
      bytes_ind->timettx_value = 0;
      break;
    case CBRBYTES:
      bytes_ind->bytes_value = htonl(gl.to_transmit);    // only meaningful if slices > 1
      bytes_ind->slices_value = htonl((u_int32_t) gl.slices);
      bytes_ind->cbr_value = htonl(gl.cbrate);
      bytes_ind->timettx_value = 0;
      break;
    case CBRTIME:
      bytes_ind->bytes_value = 0;    // only meaningful if slices > 1
      bytes_ind->slices_value = htonl((u_int32_t) gl.slices);
      bytes_ind->cbr_value = htonl(gl.cbrate);
      bytes_ind->timettx_value = htonl(gl.secmin);
      break;
    default:  // Unknown metric !
      printf("ERROR ! Unknown metric, no indication will be sent.\n");
      return 255;
  }
/*
  bytes_ind->bytes_value = htonl(gl.to_transmit);    // only meaningful if slices > 1
  bytes_ind->slices_value = htonl((u_int32_t) gl.slices);

  bytes_ind->cbr_value = htonl(gl.cbrate);
  bytes_ind->timettx_value = htonl(gl.secmin);
*/

  /* Serialization */

  /* Note : Careful with the use of sizeof(struct) for serialization or socket transmission. 
     Due to data alignment, it can be bigger than the size of the data actually allocated. 
     Padding currently disabled with #pragma pack(1), portable?
  */

  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent header: device %02x, type %04x, id_client %02x ]\n", time_to_seconds( &t0, &now ), testdef->device, ntohs(testdef->type), testdef->id_client); }
//  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent byte indication: bytes %u, slices %u ]\n", time_to_seconds( &t0, &now ), ntohl(bytes_ind->bytes_value), ntohl(bytes_ind->slices_value)); }
  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Sent byte indication: metric %s, bytes %u, slices %u, cbr %u, time-to-transmit %u ]\n", time_to_seconds( &t0, &now ), avmetrics[ntohs(bytes_ind->metric) - 1], ntohl(bytes_ind->bytes_value), ntohl(bytes_ind->slices_value), ntohl(bytes_ind->cbr_value), ntohl(bytes_ind->timettx_value)); }

  databuff = (u_char*)malloc(sizeof(struct control_header) + sizeof(struct nbr_bytes));
  if (databuff == NULL) 
  {
    fprintf(stdout, "Error allocating memory for databuff.\n");
    exit(1);
  }

  //printf("Size of malloc'd databuff : %ld = %ld + %ld (%ld)\n", sizeof(databuff), sizeof(struct control_header), sizeof(struct nbr_bytes), sizeof(u_char));

  memcpy(databuff, testdef, sizeof(struct control_header));
  memcpy(databuff + sizeof(struct control_header), bytes_ind, sizeof(struct nbr_bytes));
  free(testdef);

  //printf("Size of malloc'd databuff : %ld = %ld + %ld (%ld)\n", sizeof(databuff), sizeof(struct control_header), sizeof(struct nbr_bytes), sizeof(u_char));

  // if (gl.verbose) show_hexa(databuff, sizeof(struct control_header) + sizeof(struct nbr_bytes));

  if ((ret = send(sock_control, databuff, sizeof(struct control_header) + sizeof(struct nbr_bytes), 0)) < 0)
  {
    fprintf(stdout, "Could not send control packet, ret=%d.\n", ret);
    exit(1);
  }

  free(databuff);

  // (Implicit in the packet log)
  // if (gl.verbose) fprintf(stdout, "Control packet sent successfully (ret=%d).\n", ret);

  /* We expect now the indication of data port from the server */

  if ((testdef = receive_header(sock_control, DEV_SERVER, gl.verbose, &t0)) == NULL)
  {
    fprintf(stdout, "Error in receiving control header.\n");
    exit(1);
  }

  // is the correct type ? not checked...

  if ((ind_port = receive_dataport(sock_control)) == NULL)
  {
    fprintf(stdout, "Error in receiving data port indication.\n");
    exit(1);
  }

  if (gl.verbose) fprintf(stdout, "Indicated port for requested transmission: %d\n", ind_port->value);

  return ind_port->value;
}

int set_mptcp(int value)
{
  int mp_enabled;
  char order[BUFFSIZE];

  FILE *pp = popen("/sbin/sysctl -n net.mptcp.mptcp_enabled", "r");
  fscanf(pp, "%d\n", &mp_enabled);

  pclose(pp);

  if (gl.verbose) fprintf(stdout, "net.mptcp.mptcp_enabled=%d %s\n", value, (value == mp_enabled ? "(current)" : "" ) ); 

  if (mp_enabled != value)
  { 
    sprintf(order, "sysctl -w net.mptcp.mptcp_enabled=%d > /dev/null", value);
    if (system(order) < 0 )
    { 
      fprintf(stdout, "Error when setting mptcp_enabled parameter to value %d.\n", value);
      exit(1);
    }
  }
//  else
//  {
//    fprintf(stdout, "net.mptcp.mptcp_enabled=%d (current value)\n", mp_enabled); 
//  }

/* (In the probe, ip tcp_metrics flush brings: "tcp_metrics is unknown")
  sprintf(order, "ip tcp_metrics flush > /dev/null");
  if ((i = system(order)) < 0 )
  { 
    fprintf(stdout, "Error when flushing tcp_metrics.\n");
    exit(1);
  }
*/

}

//int test_send_cbr(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric)
int test_send_cbr(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric, int sock_ex)
{
  /* (In use) */

  /* Declarations */

  int i, ret, receive = -1;
  int sock;

  char buffer[BUFFSIZE];

  u_int32_t total_transf = 0;
  double conn_time;

  struct sockaddr_in conn;
  struct timeval time_start, time_now, time_slicestart;

  /* Set (if necessary) mptcp kernel status */

  set_mptcp(mptcp_status);

//  sock = establish(srcdataport, dstipa, dstport, nbr_if, &conn_time);
  sock = establish(srcdataport, dstipa, dstport, nbr_if, &conn_time, sock_ex);

  // And then we continue
  get_now( &time_start, 0 );

//printf("[%d] In test_send_cbr, sock %d, sock_ex %d.\n", getpid(), sock, sock_ex);

  total_transf = Sender(DEV_CLIENT, nmetric, sock, gl.to_transmit, gl.secmin, gl.cbrate, gl.slices, result);
  get_now( &time_now, 0 );

  /* Close socket */ 

/* Easy way...
  if (close(sock) != 0)
  {
    printf("ERROR, data socket %d could not be successfully closed.\n", sock);
  } 
  else {
    if (gl.debug) fprintf(stdout, "[Debug] Socket %d closed with ret 0.\n", sock);
  }
*/
// This used to be THIS complex... for some reason (in particular, when running in the probe; there were weird random, non-systematic errors... so do not discard !!!!
  if ((ret = shutdown(sock, 1)) != 0) // Values: 0 no more recv, 1 no more sending, 2 no more sending/recv (=close)
  {
    printf("ERROR, shutdown over data socket (%d) was not successful, ret=%d.\n", sock, ret);    
  }
  else if (gl.debug) fprintf(stdout, "[Debug] Shutdown socket %d with ret 0.\n", sock);

  while ((receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) > 0 )
  {
//     fprintf(stdout, "Still receiving... %d\n", receive);
  }

  if (close(sock) != 0)
  {
    printf("ERROR, data socket %d could not be successfully closed.\n", sock);
  } 
  else {
    if (gl.debug) fprintf(stdout, "[Debug] Socket %d closed with ret 0.\n", sock);
  }


  /*
  fclose(fp);  // old report file
  */

  /* Note : for a real oneway delay, substract the rtt value ? */

  /* Report */
  result[0].bytes     = total_transf;
  result[0].delay     = time_to_seconds( &time_start, &time_now );
  result[0].conn_time = conn_time;

  if (gl.slices == 1)
  {
    result[0].bandwidth = 8.0*(result[0].bytes/(result[0].delay*1000000.0));
    // result[0].bandwidth = (total_transf*8)/(result[0].delay*1000000);
  }
  else // gl.slices > 1
  {
    result[0].bandwidth = 0;
    for (i = 0; i < gl.slices; i ++)
    {
      result[0].bandwidth += result[i + 1].bandwidth/gl.slices;
    }
  }

  if (gl.verbose)
  {
    fprintf(stdout, "Total data transfered        : %ld bytes\n", result[0].bytes);
    fprintf(stdout, "Total TCP connection time    : %.8f sec\n", result[0].delay + result[0].conn_time);
    fprintf(stdout, "   Connection establishment  : %.8f sec\n", result[0].conn_time);
    fprintf(stdout, "   Transmission/release time : %.8f sec\n", result[0].delay);
    if (nmetric == BW || nmetric == BW2 || nmetric == BWALL) fprintf(stdout, "Estimated bandwidth       : %.8f Mbps\n", result[0].bandwidth);
  }

  if (result[0].bandwidth < 0) // Debug, this should not happen
  {
    fprintf(stdout, "  Warning ! Negative value for computed bandwidth\ntotal_transf %u\n  result[0].delay %lf\n", total_transf, result[0].delay);
  }

  return 0;
}


int test_recv_unif(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname, int nmetric)
{
  return test_recv(result, dstipa, dstport, nbr_if, mptcp_status, report_fname);
}

int test_recv(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname)
{
  /* Declarations */

  u_int32_t f_block_sz; // before: long (these three)
  u_int32_t transmitted = 0;
  u_int32_t total_transf = 0;
  u_int32_t tot_received = 0;

  int sock;
  struct sockaddr_in conn;
  char buffer[BUFFSIZE];
  int receive = -1, ntx = 1;

  struct tcp_info info;
  unsigned int leninfo = sizeof(info);

  struct timeval time_start, time_now;
  unsigned short opt_debug = 0;

  double conn_time;
 
  struct ifreq ifr;

  /* Set (if necessary) mptcp kernel status */

  set_mptcp(mptcp_status);

  /* Create report file */

  /*
  if (gl.verbose) fprintf(stdout, "About to create report file...\n");

  FILE *fp = fopen(report_fname, "w");
  if(fp == NULL)
  {
    fprintf(stdout, "Unable to create file %s.\n", report_fname);
    exit(1);
  }
  */

//  sock = establish(srcdataport, dstipa, dstport, nbr_if, NULL);
//  sock = establish(srcdataport, dstipa, dstport, nbr_if, NULL, 0);
  sock = establish(srcdataport, dstipa, dstport, nbr_if, &conn_time, 0);

  /*
  fprintf(fp, "    time    rto    ato snd_mss rcv_mss last-data-sent last-data-recv last-ack-recv snd-cwnd snd-ssthresh rcv-sshtresh    rtt rtt-var unacked sacked lost retrans fackets\n");
  */

  // And then we continue

  get_now( &time_start, 0 );

  // Send request
  memset(buffer, 0, BUFFSIZE);

  if(send(sock, buffer, 1, 0) < 0)
  {
    fprintf(stdout, "ERROR: Failed to send request.\n");
    exit(1);
  }

  tot_received = Receiver(DEV_CLIENT, gl.nmetric, sock, gl.to_transmit, gl.slices, result);

  get_now( &time_now, 0 );

  /* Close socket */ 
  if (close(sock) != 0)
  {
    printf("ERROR, data socket could not be successfully closed.\n");
  } 

  /*
  fclose(fp); // old report file
  */

  /* Report */
  result->bytes     = tot_received;
  result->delay     = time_to_seconds( &time_start, &time_now );
  result->bandwidth = tot_received*8/(result->delay*1000000);
  result->conn_time = conn_time;

/*
  if (gl.verbose)
  {
    fprintf(stdout, "Total data transfered     : %ld bytes\n", result->bytes);
    fprintf(stdout, "Total TCP connection time : %.8f sec\n", result->delay);
    fprintf(stdout, "Estimated bandwidth       : %.4f Mbps\n", result->bandwidth);
  }
*/

//---------
  if (gl.verbose)
  {
    fprintf(stdout, "Total data transfered        : %ld bytes\n", result->bytes);
    fprintf(stdout, "Total TCP connection time    : %.8f sec\n", result->delay + result->conn_time);
    fprintf(stdout, "   Connection establishment  : %.8f sec\n", result->conn_time);
    fprintf(stdout, "   Transmission/release time : %.8f sec\n", result->delay);
    if (gl.nmetric == BW || gl.nmetric == BW2 || gl.nmetric == BWALL) fprintf(stdout, "Estimated bandwidth       : %.8f Mbps\n", result->bandwidth);
  }
//-----------
  
  if (result->bandwidth < 0) // Debug, this should not happen
  {
    fprintf(stdout, "  Warning ! Negative value for computed bandwidth\ntotal_transf %u\n  result->delay %lf\n", total_transf, result->delay);
  }

  return 0;
}

int test_recv_unif2(struct measure * result, char* dstipa, int dstport, int nbr_if, int mptcp_status, char* report_fname)
{
  // In process (for now, it was only a copy of test_recv)
  // RECOPY from test_recv (now, it is OBSOLETE) 
  /* Declarations */

  u_int32_t f_block_sz; // before: long (these three)
  u_int32_t transmitted = 0;
  u_int32_t total_transf = 0;
  u_int32_t tot_received = 0;

  int sock;
  struct sockaddr_in conn;
  char buffer[BUFFSIZE];
  int receive = -1, ntx = 1;

  struct tcp_info info;
  unsigned int leninfo = sizeof(info);

  struct timeval time_start, time_now;
  unsigned short opt_debug = 0;

  struct ifreq ifr;

  /* Set (if necessary) mptcp kernel status */

  set_mptcp(mptcp_status);

  /* Create report file */

  /*
  if (gl.verbose) fprintf(stdout, "About to create report file...\n");

  FILE *fp = fopen(report_fname, "w");
  if(fp == NULL)
  {
    fprintf(stdout, "Unable to create file %s.\n", report_fname);
    exit(1);
  }
  */

//  sock = establish(srcdataport, dstipa, dstport, nbr_if, NULL);
  sock = establish(srcdataport, dstipa, dstport, nbr_if, NULL, 0);

  /*
  fprintf(fp, "    time    rto    ato snd_mss rcv_mss last-data-sent last-data-recv last-ack-recv snd-cwnd snd-ssthresh rcv-sshtresh    rtt rtt-var unacked sacked lost retrans fackets\n");
  */

  // And then we continue

  get_now( &time_start, opt_debug );

  // Send request
  memset(buffer, 0, BUFFSIZE);

  if(send(sock, buffer, 1, 0) < 0)
  {
    fprintf(stdout, "ERROR: Failed to send request.\n");
    exit(1);
  }

  do
  {
    memset(buffer, 0, BUFFSIZE);
    if ( (receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0 )
    {
      fprintf(stdout, "Unable to receive data from the server.\n");
      break;
    }

    tot_received+=receive;

    // Sending data acknowledgement
    if (tot_received >= ntx * gl.to_transmit / gl.slices)
    {
      memset(buffer, 0, BUFFSIZE);

      if(send(sock, buffer, 1, 0) < 0)
      {
        fprintf(stdout, "ERROR: Failed to send data acknowledgement.\n");
        exit(1);
      }

      /* Slice report */
      result[ntx].bytes     = tot_received;
      result[ntx].delay     = time_to_seconds( &time_start, &time_now );
      result[ntx].bandwidth = tot_received*8/(result->delay*1000000);
//      (result+ntx*sizeof(struct measure))->bytes     = total_transf;
//      (result+ntx*sizeof(struct measure))->delay     = time_to_seconds( &time_start, &time_now );
//      (result+ntx*sizeof(struct measure))->bandwidth = total_transf*8/(result->delay*1000000);

      ntx++;
    }
    
    get_now( &time_now, opt_debug ); // in this case, received time is ok for time_now

    /*
    if ( getsockopt(sock, 6, TCP_INFO, &info, &leninfo) < 0 ) // 6, TCP proto number
    {
      fprintf(stdout, "Unable to extract TCP_INFO struct.\n");
      exit(1);
    }	

    fprintf(fp, "%.6f %6lu %6lu %7lu %7lu %14lu %14lu %13lu %8lu %12lu %12lu %6lu %7lu %7lu %6lu %4lu %7lu %7lu\n",
			time_to_seconds( &time_start, &time_now ),
			(long unsigned int) info.tcpi_rto,
			(long unsigned int) info.tcpi_ato,
			(long unsigned int) info.tcpi_snd_mss,
			(long unsigned int) info.tcpi_rcv_mss,
			(long unsigned int) info.tcpi_last_data_sent,
			(long unsigned int) info.tcpi_last_data_recv,
			(long unsigned int) info.tcpi_last_ack_recv,
			(long unsigned int) info.tcpi_snd_cwnd,
			(long unsigned int) info.tcpi_snd_ssthresh,
			(long unsigned int) info.tcpi_rcv_ssthresh,
			(long unsigned int) info.tcpi_rtt,
			(long unsigned int) info.tcpi_rttvar,
			(long unsigned int) info.tcpi_unacked,
			(long unsigned int) info.tcpi_sacked,
			(long unsigned int) info.tcpi_lost,
	  		(long unsigned int) info.tcpi_retrans,
			(long unsigned int) info.tcpi_fackets
	  	   );
    */

    memset(buffer, 0, BUFFSIZE);
  } while (receive > 0);
/*
  if ( (receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0 )
  {
    fprintf(stdout, "Unable to receive acknowledgement from the server.\n");
    exit(1);
  }
*/
  /* Close socket */ 
  if (close(sock) != 0)
  {
    printf("ERROR, data socket could not be successfully closed.\n");
  } 

  /*
  fclose(fp);
  */

  /* Report */
  result->bytes     = tot_received;
  result->delay     = time_to_seconds( &time_start, &time_now );
  result->bandwidth = tot_received*8/(result->delay*1000000);

  if (gl.verbose)
  {
    fprintf(stdout, "Total data transfered     : %ld bytes\n", result->bytes);
    fprintf(stdout, "Total TCP connection time : %.8f sec\n", result->delay);
    fprintf(stdout, "Estimated bandwidth       : %.4f Mbps\n", result->bandwidth);
  }
  
  if (result->bandwidth < 0) // Debug, this should not happen
  {
    fprintf(stdout, "  Warning ! Negative value for computed bandwidth\ntotal_transf %u\n  result->delay %lf\n", total_transf, result->delay);
  }

  return 0;
}

struct nbr_port* receive_dataport(int sock)
{
  u_char* databuff;
  int ret;
  struct nbr_port* port_ind;

  databuff = (u_char*)malloc(sizeof(struct nbr_bytes));
  if (databuff == NULL) 
  {
    fprintf(stdout, "Error allocating memory for databuff.\n");
    exit(1);
  }

  if ( (ret=recv(sock, databuff, sizeof(struct nbr_port), 0)) < 0 )
  {
    fprintf(stdout, "Error in receiving port indication, ret=%d.\n", ret);
    exit(1);
  }

  port_ind = (struct nbr_port*)malloc(sizeof(struct nbr_port));
  if (port_ind == NULL) 
  {
    fprintf(stdout, "Error allocating memory for port_ind.\n");
    exit(1);
  }

  memcpy(port_ind, databuff, sizeof(struct nbr_port));
  free(databuff);

  // ntohs/l is required here
  port_ind->value = ntohs(port_ind->value);

  if (gl.verbose) { get_now(&now, 0); fprintf(stdout, "  [%11.5f Received port indication: value %d ]\n", time_to_seconds( &t0, &now ), port_ind->value); }

  return port_ind;
}
