#pragma pack(1)         // this avoids padding in structs, it works (at least)
                        //  in mips architecture

void parse_arguments(int device, int argc, char** argv, void * var_def, void * var)
{
  int incr = 1;

  if (device == DEV_CLIENT)
  {
    struct global_var_client * d = (struct global_var_client *) var_def;
    struct global_var_client * v = (struct global_var_client *) var;

    memcpy(var, var_def, sizeof(struct global_var_client));

    while ((argc > 1) && (argv[1][0] == '-'))
    {
      if (strlen(argv[1]) > 3) 
      {
        fprintf(stdout, "Invalid syntax: %s is not an option.\nType -h for help.\n", argv[1]);
        exit(0);
      }
 
      switch (argv[1][1])
      {
        case 'h': // Help
          fprintf(stdout, "Syntax:\n [-p <port>] [-i <dst_ipa>] [-t <bytes_to_transmit>] [-e <experiment>] [-s <nbr_slices>] [-v] [-l] [-h] [-6]\n");
          fprintf(stdout, "Parameters:\n  -p <port> : nbr of port to connect, range 1024-65535 (default %d)\n  -d <dst_ipa> : IP address of server (default %s)\n  -D : Debug mode (additional notifications) (default disabled)\n  -m <metric> : Metrics (bw bw2 bwall delay cbr-bytes cbr-time)\n  -b <cbrate> : constant bit rate (bps)\n  -t <bytes_to_transmit> : nbr of bytes to transmit (default %d)\n  -s <nbr_slices> : Number of slices (default %d)\n  -e <experiment> : Type of experiment (default %s, supported oneway reqresp)\n  -a <tag> : Tag (to be added to output lines in server)\n  -v : Verbose mode\n  -r : Collect traces\n  -l : Local model (only uses loopback interface)\n  -6 : Uses IPv6\n  -x <tx_time> : Time of transmission (in bw metric)\n  -h : Displays help\n", d->port, d->ipa, d->to_transmit, d->slices, d->test);
          // removed: \n  -k <k> : Constant k for MPTCP delay benefit metric (default %d)
          // to add: \n  -T : send tracefiles to server (default disabled)
          exit(0);
          break;
        case '6': // IPv6
          v->use6 = 1;
          incr = 1;
          break;
        case 'a': // Tag
          strcpy(v->tag, argv[2]);
          incr = 2;
          break;
        case 'b': // Constant Bit Rate
          v->cbrate = atoi(argv[2]);
          incr = 2;
          break;
        case 'd': // Destination IPA
          // sprintf(v->ipa, "%s",&argv[1][2]);
          strcpy(v->ipa, argv[2]);
          incr = 2;
          break;
        case 'D':
          v->debug = 1;
          incr = 1;
          break;
        case 'e': // Experiment (accepted values: oneway reqresp)
          strcpy(v->test, argv[2]);
          if (strcmp(v->test, "oneway") == 0) {
            v->ntest = ONEWAY;
          } else if (strcmp(v->test, "reqresp") == 0) {
            v->ntest = REQRESP;
          } else {
            fprintf(stdout, "Unknown or unsupported test: %s.\n", v->test);
            exit(1);
          } 
          incr = 2;
          break;
        case 'l': // Local (using loopback only)
          v->local=1;
          incr = 1;
          break;
        case 'm': // Metric (accepted values: bw delay)
          strcpy(v->metric, argv[2]);
          if (strcmp(v->metric, "bw") == 0) {
            v->nmetric = BW;
          } else if (strcmp(v->metric, "delay") == 0) {
            v->nmetric = DELAY;
          } else if (strcmp(v->metric, "cbr-bytes") == 0) {
            v->nmetric = CBRBYTES;
          } else if (strcmp(v->metric, "cbr-time") == 0) {
            v->nmetric = CBRTIME;
          } else if (strcmp(v->metric, "bw2") == 0) {
            v->nmetric = BW2;
          } else if (strcmp(v->metric, "bwall") == 0) {
            v->nmetric = BWALL;
          } else {
            fprintf(stdout, "Unknown or unsupported metric: %s.\n", v->metric);
            exit(1);
          } 
          incr = 2;
          break;
        case 'p': // Port
          v->port = atoi(argv[2]);
          incr = 2;
          break;
        case 'r': // Collect traces
          v->collect_traces = 1;
          incr = 1;
          break;
        case 's': // Number of slices
          v->slices = atol(argv[2]);
          incr = 2;
          break;
        case 't': // Data to transmit (bytes)
          if ((u_long) atol(argv[2]) > (u_int64_t) ((u_int64_t) 1 << (sizeof(v->to_transmit)*8)) - 1) // overflow // 1UL is "1 as unsigned long" (to avoid warnings)
//          if ((u_long) atol(&argv[1][2]) > (u_int64_t) ((u_int64_t) 1 << (sizeof(v->to_transmit)*8)) - 1) // overflow // 1UL is "1 as unsigned long" (to avoid warnings)
          {
            memset(&(v->to_transmit), 0xff, sizeof(v->to_transmit)); 
            fprintf(stdout, "Too long number of bytes (%lx > %lx), to_transmit set to maximum value (%u).\n", atol(argv[2]), (u_int64_t) ((u_int64_t) 1 << (sizeof(v->to_transmit)*8)) - 1, v->to_transmit);
          }
          else
          {
            v->to_transmit = (u_int32_t) atol(argv[2]);
          }
          incr = 2;
          break;
       /* 
        case 'k': // k constant for MPTCP Delay Benefit Metric
          v->k = atoi(argv[2]);
          incr = 2;
          break;
       */
        case 'v': // Verbose
          v->verbose=1;
          incr = 1;
          break;
        case 'x': // Time of transmission (for bw estimation)
          v->secmin = atol(argv[2]);
          incr = 2;
          break;
        default:
          fprintf(stdout, "Wrong argument: %s\nType -h for help.\n", argv[1]);
          exit(1);
      }
      argv += incr;
      argc -= incr;
    }
    
    /* Compatibility checks */
    if ((v->nmetric ==BW || v->nmetric ==BW2 || v->nmetric ==BWALL) && v->slices > 1)
    {
      fprintf(stdout, "Incompatible options: the BW metric computation cannot be performed with multiple (>1) slices.\n");
      exit(1);
    }

    var = (void *) v;
  }
  else if (device == DEV_SERVER)
  {
    struct global_var_server * d = (struct global_var_server *) var_def;
    struct global_var_server * v = (struct global_var_server *) var;

    memcpy(var, var_def, sizeof(struct global_var_client));

    while ((argc > 1) && (argv[1][0] == '-'))
    {
      if (strlen(argv[1]) > 3) 
      {
        fprintf(stdout, "Invalid syntax: %s is not an option.\nType -h for help.\n", argv[1]);
        exit(0);
      }

      switch (argv[1][1])
      {
        case 'h': // Help
          fprintf(stdout, "Syntax:\n [-p <port>] [-v] [-h] [-6]\n");
          fprintf(stdout, "Parameters:\n  -p <port> : nbr of (control) port to listen, range 1024-65535 (default %d)\n  -D <port> : nbr of (data) port to listen when requested (default %d)\n  -v : Verbose mode\n  -6 : Uses IPv6\n  -r : Collect traces\n  -h : Displays help\n", d->port, d->dataport);
          exit(0);
          break;
        case 'p': // Port
          v->port = atoi(argv[2]);
          incr = 2;
          break;
        case 'D': // Dataport
          v->dataport = atoi(argv[2]);
          incr = 2;
          break;
        case '6': // IPv6
          v->use6=1;
          strcpy(v->ipa, "::1");
          incr = 1;
          break;
        case 'r': // Collect traces
          v->collect_traces = 1;
          incr = 1;
          break;
        case 'v': // Verbose
          v->verbose=1;
          incr = 1;
          break;
        default:
          fprintf(stdout, "Wrong Argument: %s\n", argv[1]);
          fprintf(stdout, "Syntax:\n [-p<port>] [-v]\n");
          exit(1);
      }
      argv+=incr;
      argc-=incr;
    }
  }
}

void show_hexa(u_char* databuff, size_t size)
{
  /* sizeof(databuff) will return the size of the pointer, not the size of the array */

  int i;

  fprintf(stdout, "Buffer content (%ld bytes): ", size);
  for (i=0; i<size; i++) 
  {
    fprintf(stdout, "%3d ", (int)databuff[i]);
  }
  fprintf(stdout, "\n");

  return;
}

/* Get current time. */
void get_now( struct timeval *time, unsigned short debug ) 
{
  if ( gettimeofday( time, NULL ) != 0 ) {
    fprintf(stderr, "Cannot get current time.\n");
  }
  return;
}

/* Convert "struct timeval" to fractional seconds. */
double time_to_seconds ( struct timeval *tstart, struct timeval *tfinish ) 
{
  double t;

  t = (tfinish->tv_sec - tstart->tv_sec) + (tfinish->tv_usec - tstart->tv_usec) / 1e6;
  return t;
}

struct control_header* receive_header(int sock, int exp_device, int verbose, struct timeval* t0)
{
  // databuff is locally defined
  u_char* databuff;
  int ret;
  struct control_header* head;

  struct timeval now, tv;

  databuff = (u_char*)malloc(sizeof(struct control_header));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  memset(databuff, 0, sizeof(struct control_header));

  tv.tv_sec = 300;  /* 300 sec Timeout (=5 min) (60 sec is sometimes too short for 50mbit tracefile transmissions) */
  tv.tv_usec = 0;  // Not init'ing this can cause strange errors

  setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv,sizeof(struct timeval));

  if ( (ret=recv(sock, databuff, sizeof(struct control_header), 0)) <= 0 )
  {
    printf("Error in receiving control packet or no bytes received, ret=%d, errno=%d (%s).\n", ret, errno, strerror(errno));
    close(sock);
    exit(1);
//    return NULL;
  }

  // printf("  ---> received %d bytes.\n", ret);

  head = (struct control_header*)malloc(sizeof(struct control_header));
  if (head == NULL) 
  {
    printf("Error allocating memory for control_pkt.\n");
    exit(1);
  }

  memcpy(head, databuff, sizeof(struct control_header));

  head->type = ntohs(head->type);

  if (verbose) { get_now(&now, 0); printf("  [%11.5f Received header: device %02x, type %04x, id_client %02x ]\n", time_to_seconds( t0, &now ), head->device, head->type, head->id_client); }

  /* (Pending) Correctness tests for the header */
  if (head->device != exp_device) // Sending device is NOT the expected device
  {
    printf("  Warning ! Unexpected sending device from header %04x\n", head->device);
    if (verbose) show_hexa(databuff, sizeof(struct control_header) + sizeof(struct nbr_bytes));
    free(databuff);
    return NULL;
  }

  free(databuff);

  return head;  
}

long min(long a, long b)
{
  if (a > b) return b;
  return a;
}

void concatenate_string(char *original, char *add)
{
  int i, j;
  char orig[BUFFSIZE];

  strcpy(orig, original); 
  for (i=0; original[i]!='\0'; ++i);
  for (j=0; add[j]!='\0'; ++j, ++i)
  {
    original[i]=add[j];
  }
  original[i]='\0';

/*
   while(*original)
      original++;
 
   while(*add)
   {
      *original = *add;
      add++;
      original++;
   }

   *original = '\0';
*/
//printf("[Debug] First string: [%s], second string: [%s], concatenated string: [%s].\n", orig, add, original);

}

int shuffle(int * arr, int elems)
{
  int i, c, t; 

  unsigned int seed;

  FILE* random = fopen("/dev/urandom", "r");
  fread(&seed, sizeof(int), 1, random);
  fclose(random);
  srand(seed);

  for(i = 0; i < elems; i++)  // replaced "interfaces" with "av_ifs"
  {
    // int c = rand() / (RAND_MAX/elems + 1); // replaced "interfaces" with "av_ifs"
    c = rand() % elems; // (elems - 1);
    t = arr[i]; arr[i] = arr[(i+c)%elems]; arr[(i+c)%elems] = t;    // swap
    // printf("i %d , c %d\n", i, c); 
  }

  return 0;
}

pid_t launch_tcpdump(char * interf, int cliser, int id_cli, long idsession, int ntxscli, int dataport, char * tracefname)
{
//  char order[BUFFSIZE];
  char cad[10];
  pid_t pid;  

  FILE * fp;
  int traceready = 0;

  sprintf(tracefname, "%02d-%010ld-%s-%d.pcap", id_cli, idsession, (cliser == DEV_CLIENT ? "client" : "server"), ntxscli); 
  // sprintf(tracefname, "%010ld-%s-%d.pcap%s", idsession, (cliser == DEV_CLIENT ? "client" : "server"), ntxscli, (cliser == DEV_CLIENT ? ".gz" : "")); 

  pid = fork();

  if (pid < 0)
  {
    printf("Error in forking... pid=%d\n", pid);
    exit(1);          
  }
  else if (pid == 0) // Child
  {

    prctl(PR_SET_PDEATHSIG, SIGHUP); // the child dies when the parent sends SIGHUP signal (Linux-specific) // addition 25082014

    // Nota: hay que dar tiempo a que exec se realice y tcpdump se inicie (si se hace el kill antes, no hay archivo)
    sprintf(cad, "%d", dataport);

    // another possibility is to include gzip compression (-z gzip) with a limitation in the size of output files (-G)

    if (execl("/usr/sbin/tcpdump", "/usr/sbin/tcpdump", "tcp", "-n", "-i", interf, "-s", "100", "-w", tracefname, "port", cad, NULL) < 0)
    {
      printf("ERROR in exec for tcpdump.\n");
      exit(1);
    }

//    if (cliser == DEV_SERVER)
//      execl("/usr/sbin/tcpdump", "/usr/sbin/tcpdump", "tcp", "-n", "-i", interf, "-s", "100", "-w", tracefname, "port", cad, NULL);
//    else if (cliser == DEV_CLIENT)
//      execl("/usr/sbin/tcpdump", "/usr/sbin/tcpdump", "tcp", "-n", "-i", interf, "-s", "100", "-w", tracefname, "-z", "gzip", "port", cad, NULL);

  }
        
  // Parent

  while (!traceready)
  {
    //    fp = fopen(order, "r");
    fp = fopen(tracefname, "r");
    if (fp != NULL)
    {
      traceready = 1;
      fclose(fp);
    }
    sleep(1);
  }

  return pid;  
}

int stop_tcpdump(pid_t pid, int cliser, char * tracefname)
{
  int died = 0, loop;
  pid_t newpid;
  int status = 0;

  kill(pid, SIGTERM);

  for (loop = 0; !died && loop < 5; ++loop)
  {
    pid_t id;
    sleep(1);
    if (waitpid(pid, &status, WNOHANG) == pid) died = 1;
  }
  if (!died)
  {
    kill(pid, SIGKILL);
  }

  if (status != 0) return status;

  if (cliser == DEV_CLIENT) // compress
  {

    newpid = fork(); // New pid !

    if (newpid < 0)
    {
      printf("Error in forking... pid=%d\n", pid);
      exit(1);          
    }
    else if (newpid == 0) // Child
    {
      prctl(PR_SET_PDEATHSIG, SIGHUP); // the child dies when the parent sends SIGHUP signal (Linux-specific) // addition 25082014

      if (execl("/bin/gzip", "gzip", "-f", tracefname, NULL) < 0)
      {
        printf("ERROR in execl to perform gzip.\n");
        exit(1);
      }
    }
    
    // Parent
    
    waitpid(newpid, &status, 0);
    fprintf(stdout, "[Debug] Child (gzip) left with status %d.\n", status);
  }

  return status;
}

int valid_interface(struct ifreq * ifreq, int ifind, int local)
{
  int valid = TRUE;

  if (local)
  {
    if ( ! (ifreq[ifind].ifr_flags & IFF_LOOPBACK) || ! (ifreq[ifind].ifr_flags & IFF_UP) ) 
    {
//      if (gl.verbose) fprintf(stdout, "Not loopback in mode 'local': leaving...\n");
      valid = FALSE; // provisional: only use lo interface
    }
  }
  else
  {
    if ((ifreq[ifind].ifr_flags & IFF_LOOPBACK) || ! (ifreq[ifind].ifr_flags & IFF_UP)) valid = FALSE;
  }

  return valid;
}

int checkroot()
{
  int check;
  FILE *pp = popen("id -u", "r");
  fscanf(pp, "%d\n", &check);
  pclose(pp);
  return check;
}

u_int32_t Sender(int device, int nmetric, int sock, int nbytes, u_int32_t time_ttx, u_int32_t cbrate, int slices, struct measure * result)
{
  char buffer[BUFFSIZE];
  u_int32_t f_block_sz; // before: long (these three)
  u_int32_t transmitted = 0;
  // u_int32_t total_transf = 0;
  int receive = -1;

  struct timeval time_start, time_now, time_end, time_slicestart;

  int i=0, ntx=1;        // ntx is the number of slice, i is the iteration (iteration=send)
  int lasttimebytes = 0; // needed for computation of result[ntx].bytes
  int ret;

  get_now( &time_start, 0 );
  get_now( &time_slicestart, 0 );
  memset(buffer, 0, BUFFSIZE);

  if (device == DEV_SERVER) // do you really need this ? check
  {
    if ( (receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0 )
    {
      fprintf(stdout, "Unable to receive data from the client.\n");
      exit(1);
    }
  }

//  printf("[%d] In Sender. (probably dont reach here)\n", getpid());

  //printf("Sender.1: nbytes %u ; slices %d\n", nbytes, slices);

  get_now( &time_now, 0 );

  do
  {
    if (nmetric == CBRTIME || nmetric == CBRBYTES)
    {
      if (nmetric == CBRTIME)
      {
//printf("Here in CBRTIME mode %d-", time_ttx);
        if ( time_to_seconds( &time_start, &time_now ) >= time_ttx )
        {
          fprintf(stdout, "OK, you can stop.\n");
          break;
        }
      } else if (nmetric == CBRBYTES)
      {
        f_block_sz = 0;
      }

      f_block_sz = ( time_to_seconds( &time_start, &time_now ) * ((double) cbrate / 8.0) - transmitted > 1.0 ? (int) ( time_to_seconds( &time_start, &time_now ) * ((double) cbrate / 8.0) - transmitted ) : 0);  
      //printf("f_block_sz %u, cbrate: %d-", f_block_sz, cbrate);

    }
    else if (nmetric == BW || nmetric == BW2 || nmetric == BWALL)
    {
      if ( time_to_seconds( &time_start, &time_now ) >= time_ttx ) 
      {
        fprintf(stdout, "[%d] OK, you can stop.\n", getpid());
        break;
      }
      else
      {
        f_block_sz = BUFFSIZE;
      }
    } 
    else if (nmetric == DELAY)
    {
      if (nbytes > transmitted) 
      {
        if (nbytes - transmitted > min(nbytes / slices, BUFFSIZE)) f_block_sz = min(nbytes / slices, BUFFSIZE); else f_block_sz = nbytes - transmitted;
      }
      else
      {
        f_block_sz = 0; 
      }

      //printf("f_block_sz %u, cbrate: %d-", f_block_sz, cbrate);
    }

    //printf("%d %u\n", f_block_sz, cbrate); 
if (f_block_sz == 0) fprintf(stdout, "[%d] WARNING ! f_block_sz == 0, is this fine ?\n", getpid());

//printf("[%d,%d,%d-0]", getpid(), ret, transmitted);

//    if(ret = send(sock, buffer, f_block_sz, 0) < 0)
    if(ret = send(sock, buffer, f_block_sz, MSG_NOSIGNAL) < 0 && errno != EINTR)
    {
      fprintf(stdout, "ERROR: Failed to send data, iteration %d, f_block_sz=%u, transmitted %u (from %u). ret=%d (errno %s)\n", i, f_block_sz, transmitted, nbytes, ret, strerror(errno));
      break;
    }

//    if (ret == 0) 
//    {
//      fprintf(stdout, "[%d] WARNING ! send returned 0.\n", getpid()); 
//    }

    i++;
    transmitted += f_block_sz; // isn't ret the number of actually sent bytes ? No, ret is 0 if everything is ok

//printf("[%d,%d,%d-I]", getpid(), ret, transmitted);

    /* Wait until ack (if slice completed) */

    if (transmitted >= ntx*(nbytes / slices) && ( nmetric == DELAY || nmetric == CBRBYTES ) )      // if metric is BW, to_transmit has no meaning
    {
      // Data acknowledgement
      memset(buffer, 0, BUFFSIZE);
      if ( (receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0 )
      {
        fprintf(stdout, "Unable to receive data from %s.\n", (device == DEV_SERVER ? "client" : "server") );
        exit(1);
      }

      /* Slice report */
      if (ntx < slices && device == DEV_CLIENT) // the last ntx is recorded in the general result [0]
      {
        if (result != NULL) 
        {
          result[ntx].bytes     = transmitted - lasttimebytes;
          result[ntx].delay     = time_to_seconds( &time_slicestart, &time_now );
          result[ntx].bandwidth = result[ntx].bytes*8/(result[ntx].delay*1000000);
        }
        lasttimebytes = transmitted;
      }
 
      printf("  Received %d bytes (from %s) in the data socket.\n", receive, (device == DEV_SERVER ? "client" : "server") );
      ntx++;      
      get_now( &time_slicestart, 0 );
    }

//printf("[%d,%d,%d-II]", getpid(), ret, transmitted);


    get_now( &time_now, 0 ); // time_now is always computed after receiving acks to sent data

    // total_transf+=f_block_sz;

    memset(buffer, 0, BUFFSIZE);

//printf("[%d,%d,%d-III]", getpid(), ret, transmitted);


/*  ( fo not declared/opened so far )
    if (device == DEV_SERVER)
    {
      fprintf(fo, "%.6f %6lu %6lu %7lu %7lu %14lu %14lu %13lu %8lu %12lu %12lu %6lu %7lu %7lu %6lu %4lu %7lu %7lu\n",
			time_to_seconds( &time_start, &time_now ),
			(long unsigned int) info.tcpi_rto,
			(long unsigned int) info.tcpi_ato,
			(long unsigned int) info.tcpi_snd_mss,
			(long unsigned int) info.tcpi_rcv_mss,
			(long unsigned int) info.tcpi_last_data_sent,
			(long unsigned int) info.tcpi_last_data_recv,
//			(long unsigned int) info.tcpi_last_ack_sent,
			(long unsigned int) info.tcpi_last_ack_recv,
			(long unsigned int) info.tcpi_snd_cwnd,
			(long unsigned int) info.tcpi_snd_ssthresh,
			(long unsigned int) info.tcpi_rcv_ssthresh,
			(long unsigned int) info.tcpi_rtt,
			(long unsigned int) info.tcpi_rttvar,
			(long unsigned int) info.tcpi_unacked,
			(long unsigned int) info.tcpi_sacked,
			(long unsigned int) info.tcpi_lost,
	  		(long unsigned int) info.tcpi_retrans,
			(long unsigned int) info.tcpi_fackets
  	   );
    }
*/

  } while ( nmetric == DELAY || nmetric == CBRBYTES ? (transmitted < nbytes) : 1 );

  return transmitted;
}

u_int32_t Receiver(int device, int nmetric, int sock, int nbytes, int slices, struct measure * result)
//u_int32_t Receiver(int device, int nmetric, int sock, int nbytes, u_int32_t time_ttx, u_int32_t cbrate, int slices, struct measure * result)
{
  char buffer[BUFFSIZE];
  int received = -1;
  int i, ntx=1;
  u_int32_t receive, tot_received = 0;

/*
  struct tcp_info info;
  unsigned int leninfo = sizeof(info);
*/
  struct timeval time_start, time_now;

  get_now( &time_start, 0 );

  do
  {
    memset(buffer, 0, BUFFSIZE);
    if ( (receive = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0 )
    {
      fprintf(stdout, "Unable to receive data from the client.\n");
      break;
    }

    tot_received+=receive;

    // Sending data acknowledgement
    if (tot_received >= ntx * nbytes / slices && (nmetric == DELAY || nmetric == CBRBYTES) )
    {
      memset(buffer, 0, BUFFSIZE);

      if(send(sock, buffer, 1, 0) < 0)
      {
        fprintf(stdout, "ERROR: Failed to send data acknowledgement.\n");
        exit(1);
      }

      if (device == DEV_CLIENT && result != NULL)
      {
        /* Slice report */
        result[ntx].bytes     = tot_received;
        result[ntx].delay     = time_to_seconds( &time_start, &time_now );
        result[ntx].bandwidth = tot_received*8/(result->delay*1000000);
      }

      ntx++;
    }

    get_now( &time_now, 0 );

/*
    if ( getsockopt(sock, 6, TCP_INFO, &info, &leninfo) < 0 ) // 6, TCP proto number
    {
      printf("Unable to extract TCP_INFO struct.\n");
      exit(1);
    }	


    fprintf(fo, "%.6f %6lu %6lu %7lu %7lu %14lu %14lu %13lu %8lu %12lu %12lu %6lu %7lu %7lu %6lu %4lu %7lu %7lu\n",
			time_to_seconds( &time_start, &time_now ),
			(long unsigned int) info.tcpi_rto,
			(long unsigned int) info.tcpi_ato,
			(long unsigned int) info.tcpi_snd_mss,
			(long unsigned int) info.tcpi_rcv_mss,
			(long unsigned int) info.tcpi_last_data_sent,
			(long unsigned int) info.tcpi_last_data_recv,
//			(long unsigned int) info.tcpi_last_ack_sent,
			(long unsigned int) info.tcpi_last_ack_recv,
			(long unsigned int) info.tcpi_snd_cwnd,
			(long unsigned int) info.tcpi_snd_ssthresh,
			(long unsigned int) info.tcpi_rcv_ssthresh,
			(long unsigned int) info.tcpi_rtt,
			(long unsigned int) info.tcpi_rttvar,
			(long unsigned int) info.tcpi_unacked,
			(long unsigned int) info.tcpi_sacked,
			(long unsigned int) info.tcpi_lost,
	  		(long unsigned int) info.tcpi_retrans,
			(long unsigned int) info.tcpi_fackets
  	   );
*/

  } while (receive > 0);

  return tot_received;
}

double compute_bw_benefit(double mptcp_bw, double max_tcp_bw, double tot_tcp_bw)
{
  return ( mptcp_bw >= max_tcp_bw ? (mptcp_bw - max_tcp_bw)/(tot_tcp_bw - max_tcp_bw) : (mptcp_bw - max_tcp_bw)/max_tcp_bw );
}

double compute_delay_benefit(double mptcp_delay, double max_tcp_delay, double min_tcp_delay, double coeff_k)
{
  double delay_benefit;
  // double coeff_k = max_tcp_delay/min_tcp_delay;  // = Dmax/Dmin

  if (mptcp_delay <= (coeff_k/(coeff_k+1))*min_tcp_delay) {
    delay_benefit = 1;
  } else if ( (mptcp_delay > (coeff_k/(coeff_k+1))*min_tcp_delay) && (mptcp_delay <= min_tcp_delay) ) {
    delay_benefit = (coeff_k+1) * (min_tcp_delay - mptcp_delay) / (min_tcp_delay);   
//    delay_benefit = (1./coeff_k) * (min_tcp_delay - mptcp_delay) / (min_tcp_delay);   
  } else if ( (mptcp_delay > min_tcp_delay) && (mptcp_delay <= coeff_k * min_tcp_delay) ) {
    delay_benefit = (min_tcp_delay - mptcp_delay) / ((coeff_k - 1) * min_tcp_delay);    
  } else if (mptcp_delay > coeff_k * min_tcp_delay) {
    delay_benefit = -1;
  } else { // This should not happen
    fprintf(stdout, "  Warning ! Unexpected case in delay_benefit calculation: mptcp_delay %lf, min_tcp_delay %lf, max_tcp_delay %lf.\n", mptcp_delay, min_tcp_delay, max_tcp_delay);
  }

  return delay_benefit;
}

double compute_tdindex(double x, double bw_total, double max_bw)
{
  if (x >= bw_total) return 1;
  else if (x <= max_bw) return 0;
  else return (x - max_bw)/(bw_total - max_bw);
}

