#pragma pack(1)         // this avoids padding in structs, it works (at least)
                        //  in mips architecture

#include <signal.h>     // for launching and handling tcpdump
#include <sys/wait.h>   // for WNOHANG
#include <sys/prctl.h>  // for prctl (in children processes)
#include <errno.h>      // for errno
#include <stdio.h>
#include <stdlib.h>
#include <string.h>      // memset
#include <sys/socket.h>
#include <sys/ioctl.h>   // for number of interfaces
#include <net/if.h>      // for ifreq
#include <netinet/in.h>  // for sockaddr_in
#include <netinet/tcp.h> // for tcp_info
#include <unistd.h>
#include <math.h>        // for floor

#include <errno.h>

#define DEV_CLIENT 0x01
#define DEV_SERVER 0x02

#define ONEWAY     1
#define REQRESP    2

#define ONEWAYSIM  3	
#define REQRESPSIM 4

#define BW       1
#define DELAY    2
#define CBRBYTES 3
#define CBRTIME  4

#define BW2      5
#define BWALL    6 

#define TRUE    1
#define FALSE   0

#define FIVE    1000
#define ALMOST_ZERO 0.000001

#define MAXINTERFACES 20
#define NOINTERFACE   21 // bigger than MAXINTERFACES

#define CLI_ONEWAY_INSTRUCTION       0x0101
#define CLI_REQRESP_INSTRUCTION      0x0201
#define CLI_REPORT                   0x0102
#define SER_PORTINDICATION           0x0103
#define CLI_FILE                     0x0104
#define CLI_LOGLINE                  0x0105 

#define CLI_ONEWAYSIM_INSTRUCTION    0x0106
#define CLI_REQRESPSIM_INSTRUCTION   0x0107

#define CLI_ABORT                    0x01FF


#define BUFFSIZE 10000

// Define vector of constant strings, somehow
//char avmetrics = [ "bw", "delay", "cbr-bytes", "cbr-time" ];
const char * const avmetrics[] = { "bw", "delay", "cbr-bytes", "cbr-time", "bw2", "bw-all" };

struct global_var_client
{
  int verbose;
  u_int32_t to_transmit;      // to transmit or to receive (depending on the test)
  int slices;                 // number of slices
  int local;                  // for debugging in local machine (and loopback interface)
  int use6;
  char tag[20];               // max length for test tag: 10
  char test[10];              // max length for experiment name: 10
  char metric[10];            // max length for metric name: 10
  char ipa[INET6_ADDRSTRLEN]; // Set to maximum number of bytes 
                              //  (for representing the longest IPA possible, i.e. in v6)
  int port;                   // Default value, 1234

  int collect_traces;         // 1, collect; 0, do not collect

  int ntest;
  int nmetric;
  u_int32_t secmin;           // Time of transmission (for bw estimation), default 120 sec
                              // Old def: Minimum number of seconds to accept a bw estimation (hard coded)
  u_int32_t cbrate;           // Constant bit rate, for metric cbr

  int debug;      // to be removed when wlan is available in SK

  int avifs;
};

struct global_var_server
{
  int verbose;
  int use6;
  char ipa[INET6_ADDRSTRLEN];    // Set to maximum number of bytes 
                                 //  (for representing the longest IPA possible, i.e. in v6)
  int collect_traces;            // 1, collect; 0, do not collect
  int port;                      // Default value, 1234
  int dataport;                  // Default value, 5001
};

struct measure
{
  double delay;
  long   bytes;
  double bandwidth;
  double conn_time;
};

struct control_header
{
  u_char device;     // client / server
  u_short type;      // info about test and msg type
  u_char id_client;  // only meaningful if device is client
};

struct nbr_bytes
{
  u_short   metric;     // new 05.09
  u_int32_t bytes_value;
  u_int32_t slices_value;

  u_int32_t cbr_value;
  u_int32_t timettx_value;  
};

struct file_tx
{
  u_int32_t bytes;
  u_char    suffix[32];
};

struct res_benefit
{
  u_char bw_value[8];
  u_char delay_value[8];
//  u_char conntime_value[8];
};

struct nbr_port
{
  u_int16_t value;
};

void parse_arguments(int device, int argc, char** argv, void * var_def, void * var);
void show_hexa(u_char* databuff, size_t size);
void get_now( struct timeval *time, unsigned short debug );
double time_to_seconds ( struct timeval *tstart, struct timeval *tfinish );
struct control_header* receive_header(int sock, int exp_device, int verbose, struct timeval * t0);
long min(long a, long b);
void concatenate_string(char *original, char *add);
int shuffle(int * arr, int elems);
pid_t launch_tcpdump(char * interf, int cliser, int id_cli, long idsession, int ntxscli, int dataport, char * tracefname);
//pid_t launch_tcpdump(char * interf, int cliser, long idsession, int txscli, int dataport, char * tracefname);
int stop_tcpdump(pid_t pid, int cliser, char * tracefname);
int valid_interface(struct ifreq * ifreq, int ifind, int local);
int checkroot();
u_int32_t Sender(int device, int nmetric, int sock, int nbytes, u_int32_t time_ttx, u_int32_t cbrate, int slices, struct measure * result);
u_int32_t Receiver(int device, int nmetric, int sock, int nbytes, int slices, struct measure * result);
double compute_bw_benefit(double mptcp_bw, double max_tcp_bw, double tot_tcp_bw);
double compute_delay_benefit(double mptcp_delay, double max_tcp_delay, double min_tcp_delay, double coeff_k);
double compute_tdindex(double x, double bw_total, double max_bw);
