#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>       // for time_t, time() and ctime()
#include <signal.h>     // for launching and handling tcpdump
#include <unistd.h>     // for execl

#include <errno.h>      // for errno
#include <string.h>

#include <sys/select.h> // for select() in socket
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>  // for number of interfaces
#include <net/if.h>     // for ifreq
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <fcntl.h>      // for O_NONBLOCK

#include "common_meas.h" // common functions/structs
#include "common_meas.c" 

//#define BUFFSIZE 4096
#define QUEUESIZE 5     // In practice, this is NOT the maximum number of waiting clients
                        // This is protocol-, OS implementation-dependent; rather take as a lower bound
                        // (Thus, not worth turning this a CLI parameter)

#pragma pack(1)         // this avoids padding in structs, not sure it works
                        //  for any architecture (e.g. mips)

int listen_at(int port, void* parsed_conn, int npending);
struct nbr_bytes* receive_bytes(int sock);
struct res_benefit* receive_report(int sock);
int receive_file(int sock);
int receive_logline(int sock);
int send_dataport(int sock, int port);
void HandleClient(int sock, char* f_name);
void HandleSendingClient(int sock, u_int32_t nbytes);
void HandleReceivingClient(int sock, u_int32_t nbytes);
void Handle(int sock2); 

/* Command-line parameters */

struct global_var_server gl_def;
struct global_var_server gl;

/* Session parameters */
char type_test[10];
long session_bytes = 0;
short session_nmetric = 0;
int slices = 1;
unsigned char id_client = 0;
u_int32_t session_cbrate = 0;
u_int32_t session_timettx = 0;

int ini_dataport = 5001; // Default value, changeable via CLI parameter
int dataport;
time_t idsession;

struct timeval t0, time_now;

int main(int argc, char* argv[])
{
  int pid, sock, sock2;
  struct sockaddr_in conn2;
  struct sockaddr_in6 conn2_6;

  struct ifconf ifconf;
  struct ifreq ifr[50];

  int i, ifs;

  /* Default values */

  gl_def.verbose = 0;
  gl_def.use6 = 0;
  strcpy(gl_def.ipa, "127.0.0.1");
  gl_def.collect_traces = 0;
  gl_def.port = 1234;
  gl_def.dataport = ini_dataport; // i.e. 5001

  /* Force execution as root */

  if (checkroot() != 0) {
    printf("You must have root privileges to run this server.\n");
    exit(0);
  }
  
  /* Arguments parsing */ 

  parse_arguments(DEV_SERVER, argc, argv, &gl_def, &gl);

  /* Open socket to listen port */

  sock = listen_at(gl.port, (gl.use6 ? (void*) &conn2_6 : (void*) &conn2), QUEUESIZE);

/* Note: this works (to detect IPA)
//----------------------
  ifconf.ifc_buf = (char *) ifr;
  ifconf.ifc_len = sizeof ifr;

  if (ioctl(sock, SIOCGIFCONF, &ifconf) == -1) {
    printf("ERROR in ioctl\n");
    exit(1);
  }

  for (i = 0; i < ifconf.ifc_len / sizeof(ifr[0]); i++) {
    char ip[INET_ADDRSTRLEN];
    struct sockaddr_in *s_in = (struct sockaddr_in *) &ifr[i].ifr_addr;

    if (!inet_ntop(AF_INET, &s_in->sin_addr, ip, sizeof(ip))) {  // for v6 should be AF_INET6
      printf("ERROR in inet_ntop\n");
      exit(1);
    }

//&myaddr.sin_addr.s_addr

    if (strcmp(ifr[i].ifr_name, "eth0") == 0)
    {
      strcpy(gl.ipa, ip);      
      
    }
//    printf("%s - %s\n", ifr[i].ifr_name, ip);
  }
//---------------
*/

  printf("Server at %s, port %d.\n", gl.ipa, gl.port);

  unsigned int length = sizeof(conn2);
  unsigned int length6 = sizeof(conn2_6);

  get_now(&t0, 0);

  while (1) 
  {

    /* Initialize sockaddr structs */
    memset(&conn2, 0, sizeof(conn2));     // This was not the point, really... but may be useful as well (don't know)

    sock2 = ( gl.use6 ? accept(sock, (struct sockaddr*) &conn2_6, &length) : accept(sock, (struct sockaddr*) &conn2, &length) );
 
    if ( sock2 < 0 )
    {
      printf("Unable to accept connection (control channel A), error: %s.\n", strerror(errno));
      exit(1);
    }
    
    pid = fork();
    if (pid < 0)
    {
      printf("Error in forking... pid=%d\n", pid);
      exit(1);
    }

    if (pid == 0)
    {
      /* Child process */
      close(sock);

      prctl(PR_SET_PDEATHSIG, SIGHUP); // the child dies when the parent sends SIGHUP signal (Linux-specific) // addition 25082014

      /* Note : the following code is quite improvable, but does not
          matter much (could be suppressed until Handle call) */

      char * client_ipa;
      time_t now;

      if (!gl.use6)
      {
        client_ipa = (char *)malloc(INET_ADDRSTRLEN * sizeof(char));
        inet_ntop(AF_INET, &(conn2.sin_addr), client_ipa, INET_ADDRSTRLEN);
      }
      else
      {
        /* Note : in v6, the client IPA (lo) is displayed :: instead of ::1 */
        client_ipa = (char *)malloc(INET6_ADDRSTRLEN * sizeof(char));
        inet_ntop(AF_INET6, &(conn2_6.sin6_addr), client_ipa, INET6_ADDRSTRLEN);
      }

      time(&now);
      if (gl.verbose) printf("Client connected: %s, on %s", client_ipa, ctime(&now)); // \n included in ctime()

      /* IPA control: only allowed IPAs are accepted */
      if ( ( strcmp(client_ipa, "130.104.228.107") != 0 ) && 
           ( strcmp(client_ipa, "130.104.228.47") != 0 ) &&
           ( strcmp(client_ipa, "127.0.0.1") != 0 ) && 0 ) // "&& 0" turns this to FALSE; we do not have control on wifi IPAs
      {
        if (gl.verbose) printf("  Not an authorized IPA : connection refused\n");
        close(sock2);
      }
      else
      {
        Handle(sock2);
      }

      free(client_ipa);

      exit(0);
    }
    else
    {
      /* Parent process */
      //printf("[Parent] Waiting\n");
      waitpid(pid, &i, 0);
      //printf("[Parent] Now restarting the client listening loop...\n");
    }

  }

  // Close socket
  close(sock);

}

int listen_at(int port, void* parsed_conn, int npending)
{
  int sock;
  int yes=1;
  int i, j;

  // Create socket
  
  sock = ( gl.use6 ? socket(AF_INET6, SOCK_STREAM, 0) : socket(AF_INET, SOCK_STREAM, 0));

  if ( sock < 0 )
  {
    printf("Unable to create socket.\n");
    exit(1);
  }

  if (!gl.use6)
  {
    /* Note: sizeof(conn) -conn is a pointer- or sizeof(struct sockaddr_in) (or, equivalently, sizeof(*conn)) ? */ 

    struct sockaddr_in * conn = (struct sockaddr_in *) parsed_conn;
    conn = (struct sockaddr_in *) malloc(sizeof(struct sockaddr_in));
    //printf("sizeof conn %ld , sizeof *conn %ld\n", sizeof(conn), sizeof(*conn));

    memset((struct sockaddr_in*) conn, 0, sizeof(* conn));              // Clear struct
    conn->sin_family = AF_INET;                 // IPv4 address family
   // addr_n = inet_pton(AF_INET, gl.ipa, INET_ADDRSTRLEN);
    //conn->sin_addr.s_addr = htonl(addr_n); 
    conn->sin_addr.s_addr = htonl(INADDR_ANY);  // Accepting incoming packets to ANY IPA (from the host)
    conn->sin_port = htons(port);               // port (not global)


    // Setsockopt to reuse
    if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1 )
    {
      printf("Unable to configure reusing option in socket.\n");
    }

    // Bind 
    if ( bind(sock, (struct sockaddr*) conn, sizeof(*conn)) < 0 ) 
    {
      printf("Unable to bind socket to port %d, errno=%s.\n", port, strerror(errno));
      exit(1);
    }

/* Persistent binding, in principle not needed
    // Bind 
    i=-1;
    j=0;
    do  
    {
      i = bind(sock, (struct sockaddr*) conn, sizeof(*conn));
      j ++;
      printf(".");
      sleep(2);
    } while (i < 0 && j < 20);

    if (i < 0)
    {
      printf("Unable to (persistently) bind socket to port %d, errno=%d.\n", port, errno);
      exit(1);
    }
*/

  }
  else if (gl.use6) // missing the binding to IPA of the server
  {  
    struct sockaddr_in6 * conn = (struct sockaddr_in6 *) parsed_conn;
    memset(conn, 0, sizeof(* conn));              // Clear struct
    conn->sin6_family = AF_INET6;                // IPv6 address family
    conn->sin6_addr = in6addr_any;               // Incoming IPA: any
      // previous: htons(INADDR_ANY)
    conn->sin6_port = htons(port);               // port

    // Setsockopt to reuse
    if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1 )
    {
      printf("Unable to configure reusing option in socket.\n");
    }

    // Bind 
    if ( bind(sock, (struct sockaddr*) conn, sizeof(*conn)) < 0 ) 
    {
      printf("Unable to bind socket to port %d, errno=%d.\n", port, errno);
      exit(1);
    }

  }

  // Listen (queue length parsed by parameter, typically will be QUEUESIZE for control socket listening)
//  if ( listen(sock, QUEUESIZE) < 0 )   
  if ( listen(sock, npending) < 0 )   
  {
    printf("Unable to listen.\n");
    exit(1);
  }

  return sock;
}

struct nbr_bytes* receive_bytes(int sock)
{
  u_char* databuff;
  int ret;
  struct nbr_bytes* bytes_ind;

  databuff = (u_char*)malloc(sizeof(struct nbr_bytes));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  if ( (ret=recv(sock, databuff, sizeof(struct nbr_bytes), 0)) < 0 )
  {
    printf("Error in receiving control packet, ret=%d.\n", ret);
    exit(1);
  }

  bytes_ind = (struct nbr_bytes*)malloc(sizeof(struct nbr_bytes));
  if (bytes_ind == NULL) 
  {
    printf("Error allocating memory for bytes_ind.\n");
    exit(1);
  }

  memcpy(bytes_ind, databuff, sizeof(struct nbr_bytes));
  free(databuff);

  // ntohs/l is required here
  bytes_ind->metric = ntohs(bytes_ind->metric);
  bytes_ind->bytes_value = ntohl(bytes_ind->bytes_value);
  bytes_ind->slices_value = ntohl(bytes_ind->slices_value);

  bytes_ind->cbr_value = ntohl(bytes_ind->cbr_value);
  bytes_ind->timettx_value = ntohl(bytes_ind->timettx_value);

  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Received bytes indication: metric %s, bytes %u, slices %u, cbr %u, time-to-transmit %u ]\n", time_to_seconds( &t0, &time_now ), avmetrics[bytes_ind->metric - 1], bytes_ind->bytes_value, bytes_ind->slices_value, bytes_ind->cbr_value, bytes_ind->timettx_value ); }

  return bytes_ind;
}

int send_dataport(int sock, int port)
{
  u_char* databuff;
  struct control_header* testdef;
  struct nbr_port* ind_port;
  int ret;

  testdef = malloc(sizeof(struct control_header));
  if (testdef == NULL) 
  {
    printf("Error allocating memory for testdef.\n");
    exit(1);
  }

  ind_port = (struct nbr_port*)malloc(sizeof(struct nbr_port));
  if (ind_port == NULL) 
  {
    printf("Error allocating memory for ind_port.\n");
    exit(1);
  }

  testdef->device = DEV_SERVER;              // 0x01 DEV_CLIENT, 0x02 DEV_SERVER
  testdef->type = htons(SER_PORTINDICATION); // 0x0103 portind
  testdef->id_client = 0x00;                 // ignored in reception

  ind_port->value=htons(port);

  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Sent header: device %02x, type %04x ]\n", time_to_seconds( &t0, &time_now ), testdef->device, ntohs(testdef->type)); }
  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Sent data port: value %d ]\n", time_to_seconds( &t0, &time_now ), ntohs(ind_port->value)); }

  databuff = (u_char*)malloc(sizeof(struct control_header) + sizeof(struct nbr_port));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  memcpy(databuff, testdef, sizeof(struct control_header));
  memcpy(databuff + sizeof(struct control_header), ind_port, sizeof(struct nbr_port));

  free(testdef);
  free(ind_port);

  if ( (ret=send(sock, databuff, sizeof(struct nbr_port) + sizeof(struct control_header), 0)) < 0 )
  {
    printf("Error in sending port indication, ret=%d.\n", ret);
    exit(1);
  }

  free(databuff);
  
  return 0;
}

struct res_benefit* receive_report(int sock)
{
  u_char* databuff;
  int ret;
  struct res_benefit* report;

  databuff = (u_char*)malloc(sizeof(struct res_benefit));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  if ( (ret=recv(sock, databuff, sizeof(struct res_benefit), 0)) < 0 )
  {
    printf("Error in receiving control packet, ret=%d.\n", ret);
    exit(1);
  }

  report = (struct res_benefit*)malloc(sizeof(struct res_benefit));
  if (report == NULL) 
  {
    printf("Error allocating memory for report.\n");
    exit(1);
  }

  memcpy(report, databuff, sizeof(struct res_benefit));
  free(databuff);

  // No need of conversion if they are u_chars
  /* Conversion from little endian (convention) */
  // report->bw_value = le64toh(report->bw_value);
  // report->delay_value = le64toh(report->delay_value);

  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Received report: bw benefit %s, delay benefit %s ]\n", time_to_seconds( &t0, &time_now ), report->bw_value, report->delay_value); }

  return report;
}

int receive_file(int sock)
{
  char buffer[BUFFSIZE];
  u_char* databuff;
  u_int32_t ret, received = 0;
  struct file_tx* file_charact;

  databuff = (u_char*)malloc(sizeof(struct file_tx));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  if ( (ret=recv(sock, databuff, sizeof(struct file_tx), 0)) < 0 )
  {
    printf("Error in receiving control packet, ret=%d.\n", ret);
    exit(1);
  }

  file_charact = (struct file_tx*)malloc(sizeof(struct file_tx));
  if (file_charact == NULL) 
  {
    printf("Error allocating memory for file characteristics.\n");
    exit(1);
  }

  memcpy(file_charact, databuff, sizeof(struct file_tx));

  file_charact->bytes = ntohl(file_charact->bytes);

  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Received file characteristics: bytes %d, suffix %s ]\n", time_to_seconds( &t0, &time_now ), file_charact->bytes, file_charact->suffix); }

  /* Format : <id_client>-<secs_from_1_1_1970>-<suffix>.txt */

  char filename[BUFFSIZE], addition[BUFFSIZE];

  sprintf(filename, "%02x-%010ld-%s", id_client, idsession, file_charact->suffix);

  FILE * fp = fopen(filename, "w");

  do 
  {
    memset(buffer, 0, BUFFSIZE);
//    if ( (ret = (u_int32_t) recv(sock, buffer, sizeof(buffer), 0)) < 0)
    if ( (ret = (u_int32_t) recv(sock, buffer, 1, 0)) < 0)
    {
      printf("Unable to receive file on the control channel from client.\n");
      break;
    }

    fwrite(buffer, sizeof(char), ret, fp);
//    fprintf(fp, "%s", buffer);

    received += ret;

  } while (received < file_charact->bytes);

  fclose(fp);

  printf("  OK, received %d bytes.\n", received);


  free(databuff);

  return 0;
}

int receive_logline(int sock)
{
  char buffer[BUFFSIZE];
  u_char* databuff;
  u_int32_t ret, received = 0;
  struct file_tx* file_charact;

  databuff = (u_char*)malloc(sizeof(struct file_tx));
  if (databuff == NULL) 
  {
    printf("Error allocating memory for databuff.\n");
    exit(1);
  }

  if ( (ret=recv(sock, databuff, sizeof(struct file_tx), 0)) < 0 )
  {
    printf("Error in receiving control packet, ret=%d.\n", ret);
    exit(1);
  }

  file_charact = (struct file_tx*)malloc(sizeof(struct file_tx));
  if (file_charact == NULL) 
  {
    printf("Error allocating memory for file characteristics.\n");
    exit(1);
  }

  memcpy(file_charact, databuff, sizeof(struct file_tx));

  file_charact->bytes = ntohl(file_charact->bytes);

  if (gl.verbose) { get_now(&time_now, 0); printf("  [%11.5f Received log line characteristics: bytes %d ]\n", time_to_seconds( &t0, &time_now ), file_charact->bytes); }

  char filename[BUFFSIZE], addition[BUFFSIZE];

  strcpy(filename, file_charact->suffix);

  FILE * fp = fopen(filename, "a");

  // New Aug 7 -- addition of session id in logfile
  sprintf(buffer, "%010ld ", idsession);
  fwrite(buffer, sizeof(char), 11, fp);  

  do 
  {
    memset(buffer, 0, BUFFSIZE);
    if ( (ret = (u_int32_t) recv(sock, buffer, 1, 0)) < 0)
    {
      printf("Unable to receive file on the control channel from client.\n");
      break;
    }

/*int i;
for (i=0; i<ret; i++)
{
printf("[%c]", buffer);
}
*/
    fwrite(buffer, sizeof(char), ret, fp);

    received += ret;

    //printf("%d-", received);

  } while (received < file_charact->bytes);

  // Adding CR (\n) in the end
  fwrite("\n", sizeof(char), 1, fp);

  fclose(fp);

  printf("  OK, received %d bytes.\n", received);

  free(databuff);

  return 0;
}


void Handle(int sock2)
{

        fd_set active_set, reading_set;

        int reczero[MAXINTERFACES], maxreczero=5;

        int i;
        u_char buffer[BUFFSIZE], receive;
        int active_sockets = 0;
        int client_socket[MAXINTERFACES], ntx[MAXINTERFACES], sd, max_sd = 0;
        u_int32_t tot_received[MAXINTERFACES];
        int nchildren = 0, pid_sc[MAXINTERFACES], status; // fork for REQRESPSIM 201114
        for (i=0; i<MAXINTERFACES; i++) { client_socket[i]=0; reczero[i]=0; ntx[i]=1; tot_received[i]=0; pid_sc[i]=0;}

        struct sigaction sa;





  struct sockaddr_in conn2;
  struct sockaddr_in6 conn2_6;

  struct nbr_port *ind_port;
  struct control_header *testdef;
  struct control_header* control_pkt;
  struct nbr_bytes* bytes_ind;
  struct res_benefit* report;
  int ret = 0;
  int terminate = 0;

  int ntxscli = 0;

  char order[BUFFSIZE];
  pid_t pid;

  unsigned int length;

  struct timeval timeout;
  int maxwaiting = 20; // should be command-line variable

  struct timeval session_start, session_end;

  fd_set master_set, working_set; // without struct !
  int rc;

  int sock_data, sock_data2;

  u_char* databuff;

  get_now(&session_start, 0);

  // if (gl.verbose) fprintf(stdout, "Starting session...\n");

  while (! terminate)
  {
    if ((control_pkt = receive_header(sock2, DEV_CLIENT, gl.verbose, &t0)) == NULL)
    {
      printf("Error in receiving header.\n");
      terminate = 1;
      continue;
      // exit(1);
    }

    if (id_client == 0) // first packet received in the session
    {
      // Session initialization
      id_client = control_pkt->id_client;
      dataport = gl.dataport; //ini_dataport;
      idsession = time(NULL);
      if (gl.verbose) printf("  Id client: %02d ; Id session: %ld\n", id_client, idsession); 
      if (id_client == 255)
      {
        printf("  WARNING ! Not proper client identification, client used default value 255.\n");
      }
    }
    else if (id_client != control_pkt->id_client) // client mismatch, abort session
    {
      printf("  ERROR : Unexpected client id (current id: %d, id received: %d), aborting...\n", id_client, control_pkt->id_client);
      terminate = 1;
      continue;
    }

    // if (gl.verbose) printf("  Control packet received: device %02x, type %04x\n", (int)control_pkt->device, (int)control_pkt->type);

    switch(control_pkt->type)
    {
      case CLI_ABORT:                  // client order: abort

        printf("  Termination forced by the client, something went wrong...\n");
        terminate = 1;

        /* Initialize session variables */

        session_bytes = 0;
        session_nmetric = 0;
        
        break;
      case CLI_REQRESPSIM_INSTRUCTION: // instruction from client 0x0107
        // Originally, a copy from ONEWAYSIM_INSTRUCTION
        maxreczero=5;

        active_sockets = 0;
        max_sd = 0;
        int nchildren = 0; // fork for REQRESPSIM 201114
        for (i=0; i<MAXINTERFACES; i++) { client_socket[i]=0; reczero[i]=0; ntx[i]=1; tot_received[i]=0; pid_sc[i]=0;}

        /* Get number of bytes */
        ntxscli++;

        // if (gl.verbose) printf("  Request/response test (sending indication)\n");
        strcpy(type_test, "reqresp");

        if ((bytes_ind = receive_bytes(sock2)) == NULL)
        {
          printf("Error in receiving number of bytes indication.\n");
          exit(1);
        }

        /* Check bytes_ind consistency (if needed) */

        if (session_bytes > 0 && bytes_ind->bytes_value != session_bytes)
        {
          printf("  Warning, inconsistent number of bytes ! Overriding old value of session_bytes...\n");
        }
        session_bytes = bytes_ind->bytes_value;
        session_nmetric = bytes_ind->metric;
        slices = bytes_ind->slices_value;            // this is the number of interfaces at the MA
        session_cbrate = bytes_ind->cbr_value;
        session_timettx = bytes_ind->timettx_value;

        // Initialize sockaddr struct
        memset(&conn2, 0, sizeof(conn2)); 

        // Listen at port 5001
        sock_data = listen_at(dataport, (gl.use6 ? (void*) &conn2_6 : (void*) &conn2), MAXINTERFACES);

        printf("  Now listening at port %d.\n", dataport);

        // No trace collection for now (REQRESPSIM)
        //if (gl.collect_traces) pid = launch_tcpdump("eth0", DEV_SERVER, id_client, idsession, ntxscli, dataport, order);

        // Now it's time to indicate port to connect
        if ((ret = send_dataport(sock2, dataport)) < 0)
        {
          printf("Error sending port indication.\n");

          // Thus, close data socket (nobody will write there anyway)

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 
          //close(sock_data)

          // No trace collection for now (REQRESPSIM)
          //if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          exit(1);
        }

        /* Note : if dst port is kept constant, then enabling SO_REUSEADDR is recommended */

        //dataport++;


 //   /* Set up the signal handler */
 //   sa.sa_handler = wait_for_child;
 //   sigemptyset(&sa.sa_mask);
 //   sa.sa_flags = SA_RESTART;
 //   if (sigaction(SIGCHLD, &sa, NULL) == -1) {
 //      perror("sigaction");
 //       return 1;
 //   }

        FD_ZERO(&active_set);
        FD_ZERO(&reading_set);

        FD_SET(sock_data, &active_set);
        active_sockets = 1;
        max_sd = sock_data;

printf("Number of interfaces at MA: %d\n", slices);

        do { /* Una posibilidad es fijar el timeout at tiempo que se tiene que transmitir (pero es feo; cerraras abruptamente y no sabes que pasara con los sockets...) */
          timeout.tv_sec  = session_timettx; //maxwaiting; // in sec
          timeout.tv_usec = 0;          // usec < 1 sec

          FD_ZERO(&reading_set); // new 13/11/2014 (not necessary, apparently)

          memcpy(&reading_set, &active_set, sizeof(active_set)); // new 07/11/2014

          rc = select(max_sd+1, &reading_set, NULL, NULL, &timeout);  // you need active_set and reading_set because select CHANGES the fd set it is parsed
//          printf("(%d)", rc);
          if (rc < 0)
          {
            printf("Error in select (%s).\n", strerror(errno));

            if (shutdown(sock_data, 1) != 0)
            {
              printf("ERROR in shutting down data socket sock_data.\n");
            } 
            while(recv(sock_data, NULL, 0, 0) == 0); 

            // No traces for now...
            // if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
            if (close(sock_data) != 0)
            {
              printf("ERROR in closing data socket sock_data (%s).\n", strerror(errno));
            }

            exit(1);
          }
          else if (rc == 0) 
          {
            printf("  Timeout (%d sec): no connections received at data port.\n", maxwaiting);

            //Properly close socket : shutdown(sock, 1) , 2) while(read(sock, ...)=0) 3) close(sock)

            if (shutdown(sock_data, 1) != 0)
            {
              printf("ERROR in shutting down data socket sock_data.\n");
            } 
            while(recv(sock_data, NULL, 0, 0) == 0); 

            // No trace collection so far
            //if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
            if (close(sock_data) != 0)
            {
              printf("ERROR in closing data socket sock_data.\n");
            }

            terminate = 1;
            break;
          }
          else if (rc > 0 && FD_ISSET(sock_data, &reading_set)) // new
          {
            length = sizeof(conn2);

            if ((sock_data2 = accept(sock_data, (struct sockaddr*) &conn2, &length)) < 0)
            {
              printf("Unable to accept connection (data channel C), error %s.\n", strerror(errno));
              exit(1);
            }

            nchildren ++;
            pid_sc[nchildren - 1] = fork();
            if (pid_sc[nchildren - 1] == 0) {
              /* In child process */
              close(sock_data);
              int total_transf;
              total_transf = Sender(DEV_SERVER, session_nmetric, sock_data2, session_bytes, session_timettx, session_cbrate, slices, NULL);
              exit(0);
            }
            else {
              /* Parent process */
              if (pid_sc[nchildren - 1] == -1) {
                perror("fork");
                exit(1);
              }
              else {
                close(sock_data2);
              }
              if (nchildren >= slices) // all subflows have been created
                break;
            }

          }

        } while (1); // the listener is always there

        for (i=0; i<nchildren; i++)
        {
          // Wait for children
          printf("Waiting for %d...\n", pid_sc[i]);

          if (waitpid(pid_sc[i], &status, 0) > 0)
          {
            printf("  (Child %d (pid %d) terminated.)\n", i, pid_sc[i]);
            if (status != 0)
            {
              fprintf(stdout, "Not-so-normal termination of child %i (pid %d):", i, pid_sc[i]);
              if (WIFSTOPPED(status))   fprintf(stdout, "  WIFSTOPPED\n");
              if (WIFCONTINUED(status)) fprintf(stdout, "  WIFCONTINUED\n");
              if (WIFSIGNALED(status))  fprintf(stdout, "  WIFSIGNALED (signal %d)\n", WTERMSIG(status));
              if (WIFEXITED(status))    fprintf(stdout, "  WIFEXITED (signal %d)\n", WEXITSTATUS(status));
            }
          }
          else
          {
            fprintf(stdout, "ERROR when waiting for child %d (pid %d): %s.\n", i, pid_sc[i], strerror(errno));
          }
        
        }

//        }

        sleep(1); // wait to "ensure" that socket is really closed

        close(sock_data);

        break;


      case CLI_ONEWAYSIM_INSTRUCTION:  // instruction from client 0x0106
      // Originally, a copy of case CLI_ONEWAY_INSTRUCTION (below)
 
        /* Get number of bytes */
        ntxscli++;

        // if (gl.verbose) printf("  One-way test (sending indication)\n");
        strcpy(type_test, "oneway");

        if ((bytes_ind = receive_bytes(sock2)) == NULL)
        {
          printf("Error in receiving number of bytes indication.\n");
          exit(1);
        }

        /* Check bytes_ind consistency (if needed) */

        if (session_bytes > 0 && bytes_ind->bytes_value != session_bytes)
        {
          printf("  Warning, inconsistent number of bytes ! Overriding old value of session_bytes...\n");
        }
        session_bytes = bytes_ind->bytes_value;
        session_nmetric = bytes_ind->metric;
        slices = bytes_ind->slices_value;
        session_cbrate = bytes_ind->cbr_value;
        session_timettx = bytes_ind->timettx_value;

        if (gl.verbose) if (slices > 1) printf("  Expecting %u bytes...\n", bytes_ind->bytes_value);

        // Initialize sockaddr struct
        memset(&conn2, 0, sizeof(conn2)); 

        // Listen at port 5001
        sock_data = listen_at(dataport, (gl.use6 ? (void*) &conn2_6 : (void*) &conn2), MAXINTERFACES);

        printf("  Now listening at port %d.\n", dataport);

        // No trace collection for now (ONEWAYSIM)
        //if (gl.collect_traces) pid = launch_tcpdump("eth0", DEV_SERVER, id_client, idsession, ntxscli, dataport, order);

        // Now it's time to indicate port to connect
        if ((ret = send_dataport(sock2, dataport)) < 0)
        {
          printf("Error sending port indication.\n");

          // Thus, close data socket (nobody will write there anyway)

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 
          //close(sock_data)

          // No trace collection for now (ONEWAYSIM)
          //if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          exit(1);
        }

        /* Note : if dst port is kept constant, then enabling SO_REUSEADDR is recommended */

        //dataport++;

        maxreczero=5;

        active_sockets = 0;
        max_sd = 0;
        nchildren = 0; // fork for REQRESPSIM 201114
        for (i=0; i<MAXINTERFACES; i++) { client_socket[i]=0; reczero[i]=0; ntx[i]=1; tot_received[i]=0; pid_sc[i]=0;}

        FD_ZERO(&active_set);
        FD_ZERO(&reading_set);

        FD_SET(sock_data, &active_set);
        active_sockets = 1;
        max_sd = sock_data;


        do {
          timeout.tv_sec  = maxwaiting; // in sec
          timeout.tv_usec = 0;          // usec < 1 sec

          FD_ZERO(&reading_set); // new 13/11/2014 (not necessary, apparently)

          memcpy(&reading_set, &active_set, sizeof(active_set)); // new 07/11/2014

        rc = select(max_sd+1, &reading_set, NULL, NULL, &timeout);  // you need active_set and reading_set because select CHANGES the fd set it is parsed
//        printf("(%d)", rc);
        if (rc < 0)
        {
          printf("Error in select (%s).\n", strerror(errno));

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 

          // No traces for now...
          // if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data (%s).\n", strerror(errno));
          }

          exit(1);
        }
        else if (rc == 0) 
        {
          printf("  Timeout (%d sec): no connections received at data port.\n", maxwaiting);

          //Properly close socket : shutdown(sock, 1) , 2) while(read(sock, ...)=0) 3) close(sock)

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 

          // No trace collection so far
          //if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          terminate = 1;
          break;
        }
        else if (rc > 0 && FD_ISSET(sock_data, &reading_set)) // new
        {
          length = sizeof(conn2);

          if ((sock_data2 = accept(sock_data, (struct sockaddr*) &conn2, &length)) < 0)
          {
            printf("Unable to accept connection (data channel D), error: %s.\n", strerror(errno));
            exit(1);
          }

          //printf("(rc=%d) *", rc);

          //inform user of socket number - used in send and receive commands
          //printf("New connection , socket fd is %d , ip is : %s , port : %d \n" , new_socket , inet_ntoa(address.sin_addr) , ntohs(address.sin_port));
         
//          fcntl(sock_data2, F_SETFL, O_NONBLOCK); // new 051114

//int buffsize = 65365;
//setsockopt(sock_data2, SOL_SOCKET, SO_RCVBUF, &buffsize, sizeof(buffsize));
            
          //add new socket to array of sockets
          for (i = 0; i < MAXINTERFACES; i++)
          {
            //if position is empty
            if( client_socket[i] == 0 )
            {
              client_socket[i] = sock_data2;
              FD_SET(client_socket[i], &active_set);
              if (client_socket[i] > max_sd) max_sd = client_socket[i];
              active_sockets ++;                   
              printf("Created new data socket %d, max_sd %d \n", sock_data2, max_sd);
              break;
            }
          }
        }
        else if (rc > 0 && ! FD_ISSET(sock_data, &reading_set)) // disabling this doesnt seem to help
        {
          // Otherwise event in select corresponds to one of the child sockets

          for (i=0; i < MAXINTERFACES; i++)
          {
            if (client_socket[i] != 0 && FD_ISSET(client_socket[i], &reading_set))
            {
              /* In here we need to behave as Receiver (ONEWAYSIM) and as Sender (REQRESPSIM) */

              // OK, here we need to receive data from the corresponding client (replicated code from Receiver)

              memset(buffer, 0, BUFFSIZE);
              receive = (u_int32_t) recv(client_socket[i], buffer, sizeof(buffer), 0);
//              receive = (u_int32_t) recv(client_socket[i], buffer, sizeof(buffer), MSG_DONTWAIT);
              if ( receive < 0 )
              {
                fprintf(stdout, "Unable to receive data from the client %d.\n", i);

                close(client_socket[i]);
                FD_CLR(client_socket[i], &active_set);
                client_socket[i] = 0;
                // also, active_sockets -- (but if there is an error, who cares ?)

                exit(1);
                // handle error, what to do now ? should leave...
              }
              else if (receive > 0)
              {
                printf("%d", i);
                reczero[i] = 0; // reinitilize reczero counter

                tot_received[i]+=receive;
               /*
                // Sending data acknowledgement
                if (tot_received[i] >= ntx[i] * session_bytes / slices && (session_nmetric == DELAY || session_nmetric == CBRBYTES) )
                {
                  memset(buffer, 0, BUFFSIZE);

                  if(send(client_socket[i], buffer, 1, 0) < 0)
                  {
                    fprintf(stdout, "ERROR: Failed to send data acknowledgement.\n");
                    exit(1);
                  }

                  printf("Send ack over data socket %d ! \n", i);
                  ntx[i]++;
                } // endif (ack is needed)
               */
              } // endif (receive > 0)
              else if (receive == 0)
              {
               reczero[i]++;
               printf("(%d)", reczero[i]);
//               printf("recv=0 in data socket %d (reczero=%d, rc=%d, ISSET(sock_data) %d, ISSET(client_socket[i]) %d, receive=%d).\n", i, reczero[i], rc, FD_ISSET(sock_data, &reading_set), FD_ISSET(client_socket[i], &reading_set), receive);

               if (reczero[i] > maxreczero) {
                // printf("Data socket %d closed by client %d (reczero=%d, rc=%d, ISSET(sock_data) %d, ISSET(client_socket[i]) %d, receive=%d) (errno %s).\n", client_socket[i], i, reczero[i], rc, FD_ISSET(sock_data, &reading_set), FD_ISSET(client_socket[i], &reading_set), receive, strerror(errno));
                FD_CLR(client_socket[i], &active_set);
                //      close(client_socket[i]); // sockets will be closed later
                //redundant ? necessary ?
                // client_socket[i] = 0;
                active_sockets --;
               }
              }
 
              break;

            } // endif FD_ISSET
          } // endfor

        }
        else
        {
          printf("ERROR ! This is not supposed to happen, case not covered\n"); // SHould never happen
        }

        } while (active_sockets > 1); // the listener is always there

/*
          pid_server = fork();
          switch(pid_server)
          {
            case 0: // child

              length = sizeof(conn2);
              if ( (sock_data2 = accept(sock_data, (struct sockaddr*) &conn2, &length)) < 0 )
              {
                printf("Unable to accept connection.\n");
                exit(1);
              }

              HandleSendingClient(sock_data2, bytes_ind->bytes_value);

              if (close(sock_data2) != 0)
              {
                printf("ERROR in closing data socket sock_data2.\n");
              }         
              
              exit(0);

              break;
            case -1: // error
              fprintf(stdout, "ERROR in fork\n");
              break;
        //    default: // parent
          }
*/          
        //}

        sleep(1); // wait to "ensure" that socket is really closed

        // Why ? because they are not closed before
        for (i=0; i<MAXINTERFACES; i++)
        {
          if (client_socket[i] != 0) {

      if (control_pkt->type == CLI_REQRESPSIM_INSTRUCTION)
      {
      // Wait for children
        printf("Waiting for %d...\n", pid_sc[i]);

        if (waitpid(pid_sc[i], &status, 0) > 0)
        {
          printf("  (Child %d (pid %d) terminated.)\n", i, pid_sc[i]);
          if (status != 0)
          {
            fprintf(stdout, "Not-so-normal termination of child %i (pid %d):", nchildren - 1, pid_sc[nchildren - 1]);
            if (WIFSTOPPED(status))   fprintf(stdout, "  WIFSTOPPED\n");
            if (WIFCONTINUED(status)) fprintf(stdout, "  WIFCONTINUED\n");
            if (WIFSIGNALED(status))  fprintf(stdout, "  WIFSIGNALED (signal %d)\n", WTERMSIG(status));
            if (WIFEXITED(status))    fprintf(stdout, "  WIFEXITED (signal %d)\n", WEXITSTATUS(status));
          }
          nchildren --;
        }
        else
        {
          fprintf(stdout, "ERROR when waiting for child %d (pid %d): %s.\n", nchildren - 1, pid_sc[nchildren - 1], strerror(errno));
        }
      }


            close(client_socket[i]);
          }
        }

        // No trace collection for now (ONEWAYSIM)
        //if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);

        close(sock_data);

       // ----------------- addition 09/04/2014
       // exit(0);
       // }
       // ----------------- end addition 09/04/2014

        break;

      case CLI_ONEWAY_INSTRUCTION: // instruction from client 0x0101 
        /* Get number of bytes */
        ntxscli++;

        // if (gl.verbose) printf("  One-way test (sending indication)\n");
        strcpy(type_test, "oneway");

        if ((bytes_ind = receive_bytes(sock2)) == NULL)
        {
          printf("Error in receiving number of bytes indication.\n");
          exit(1);
        }

        /* Check bytes_ind consistency (if needed) */

        if (session_bytes > 0 && bytes_ind->bytes_value != session_bytes)
        {
          printf("  Warning, inconsistent number of bytes ! Overriding old value of session_bytes...\n");
        }
        session_bytes = bytes_ind->bytes_value;
        session_nmetric = bytes_ind->metric;
        slices = bytes_ind->slices_value;
        session_cbrate = bytes_ind->cbr_value;
        session_timettx = bytes_ind->timettx_value;

        if (gl.verbose) if (slices > 1) printf("  Expecting %u bytes...\n", bytes_ind->bytes_value);

        // Initialize sockaddr struct
        memset(&conn2, 0, sizeof(conn2)); // OK, this didn't fix the problem

        // Listen at port 5001
        sock_data = listen_at(dataport, (gl.use6 ? (void*) &conn2_6 : (void*) &conn2), MAXINTERFACES);

        printf("  Now listening at port %d.\n", dataport);

        // if (gl.collect_traces) pid = launch_tcpdump("eth0", DEV_SERVER, idsession, ntxscli, dataport, order);
        if (gl.collect_traces) pid = launch_tcpdump("eth0", DEV_SERVER, id_client, idsession, ntxscli, dataport, order);

        // Now it's time to indicate port to connect

        if ((ret = send_dataport(sock2, dataport)) < 0)
        {
          printf("Error sending port indication.\n");

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 
          //close(sock_data)

          if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          exit(1);
        }

        /* Note : if dst port is kept constant, then enabling SO_REUSEADDR is recommended */

        //dataport++;

        FD_ZERO(&master_set);
        FD_SET(sock_data, &master_set);
 
        timeout.tv_sec  = maxwaiting; // in sec
        timeout.tv_usec = 0;

        memcpy(&working_set, &master_set, sizeof(master_set));
        
        rc = select(sock_data+1, &working_set, NULL, NULL, &timeout);
        if (rc < 0)
        {
          printf("Error in select.\n");

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 

          if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          exit(1);
        }
        else if (rc == 0)
        {
          printf("  Timeout (%d sec): no connections received at data port.\n", maxwaiting);

          //Properly close socket : shutdown(sock, 1) , 2) while(read(sock, ...)=0) 3) close(sock)

          if (shutdown(sock_data, 1) != 0)
          {
            printf("ERROR in shutting down data socket sock_data.\n");
          } 
          while(recv(sock_data, NULL, 0, 0) == 0); 

          if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);
          
          if (close(sock_data) != 0)
          {
            printf("ERROR in closing data socket sock_data.\n");
          }

          terminate = 1;
          break;
        }

        length = sizeof(conn2);
        if ( (sock_data2 = accept(sock_data, (struct sockaddr*) &conn2, &length)) < 0 )
        {
          printf("Unable to accept connection (data channel A), error: %s.\n", strerror(errno));
          exit(1);
        }

        HandleSendingClient(sock_data2, bytes_ind->bytes_value);

        if (close(sock_data2) != 0)
        {
          printf("ERROR in closing data socket sock_data2.\n");
        }

        sleep(1); // wait to "ensure" that socket is really closed

        if (gl.collect_traces) stop_tcpdump(pid, DEV_SERVER, NULL);

        close(sock_data);

       // ----------------- addition 09/04/2014
       // exit(0);
       // }
       // ----------------- end addition 09/04/2014

        break;

      case CLI_REQRESP_INSTRUCTION: // instruction from client (req/resp) 0x0201

        // if (gl.verbose) printf("  Request/response test (sending indication)\n");
        strcpy(type_test, "reqresp");

        /* Get number of bytes */
        if ((bytes_ind = receive_bytes(sock2)) == NULL)
        {
          printf("Error in receiving number of bytes indication.\n");
          exit(1);
        }

        /* Check bytes_ind consistency (if needed) */

        if (session_bytes > 0 && bytes_ind->bytes_value != session_bytes)
        {
          printf("  Warning, inconsistent number of bytes ! Overriding old value of session_bytes...\n");
        }
        session_bytes = bytes_ind->bytes_value;
        session_nmetric = bytes_ind->metric;
        slices = bytes_ind->slices_value;
        session_cbrate = bytes_ind->cbr_value;
        session_timettx = bytes_ind->timettx_value;

        if (gl.verbose) printf("  Ready to send %u bytes...\n", bytes_ind->bytes_value);

        /* Note: check all this part, should be similar to oneway */

        // Listen at port 5001
        sock_data = listen_at(dataport, (gl.use6 ? (void*) &conn2_6 : (void*) &conn2), MAXINTERFACES);

        printf("  Now also listening at port %d.\n", dataport);


        // Now it's time to indicate port to connect

        if ((ret = send_dataport(sock2, dataport)) < 0)
        {
          printf("Error sending port indication.\n");
          exit(1);
        }

        /* Note : if dst port is kept constant, then enabling SO_REUSEADDR is recommended */

        // dataport++;

        FD_ZERO(&master_set);
        FD_SET(sock_data, &master_set);
 
        timeout.tv_sec  = maxwaiting; // in sec
        timeout.tv_usec = 0;

        memcpy(&working_set, &master_set, sizeof(master_set));
        
        rc = select(sock_data+1, &working_set, NULL, NULL, &timeout);
        if (rc < 0)
        {
          printf("Error in select.\n");
          exit(1);
        }

        if (rc == 0)
        {
          printf("  Timeout (%d sec): no connections received at data port.\n", maxwaiting);
          close(sock_data);
          terminate = 1;
          break;
        }

        length = sizeof(conn2);
        if ( (sock_data2 = accept(sock_data, (struct sockaddr*) &conn2, &length)) < 0 )
        {
          printf("Unable to accept connection (data channel B), error: %s.\n", strerror(errno));
          exit(1);
        }

        HandleReceivingClient(sock_data2, bytes_ind->bytes_value);

        close(sock_data2);
        close(sock_data);

        break;

      case CLI_FILE: // file transmitted from client 0x0104

        if ( receive_file(sock2) != 0 )
        {
          printf("Error in receiving client file.\n");
          exit(1);
        }
    
        break;

      case CLI_LOGLINE: // line to be appended to log file, transmitted from client 0x0105

        if ( receive_logline(sock2) != 0 )
        {
          printf("Error in receiving client file.\n");
          exit(1);
        }
    
        break;

      case CLI_REPORT: // report from client 0x0102

        /* Note: the reception of a "report" signalizes the end of a "session",
                 in each "session" there may be several "indications" received,
                 but each "indication" needs to be consistent (test and bytes).   */

        // if (gl.verbose) printf("  Report\n");
/* Only one type of report so far
        if (gl.verbose) printf("  One-way test (report)\n");

        if (strcmp(type_test, "oneway") != 0) // mismatch
        {
          printf("Report from an unexpected test type.\n");
          exit(1);
        }
*/
        if ((report = receive_report(sock2)) == NULL)
        {
          printf("Error in receiving client report.\n");
          exit(1);
        }

        /* Log in file */

        get_now( &time_now, 0 );

        FILE *fplog = fopen("log.txt", "a");
        if (fplog == NULL)
        {
          printf("Unable to write on log file.\n");
        }
        else 
        {
          fprintf(fplog, "%11.5f probe%03d %10s %10ld %3d %11s %11s\n", time_to_seconds( &t0, &time_now ), (int) id_client, type_test, session_bytes, slices, report->bw_value, report->delay_value);
          fclose(fplog);
        }

        if (gl.verbose) 
        {
          printf("  MPTCP aggregate BW benefit, within -1 and 1: %s\n", report->bw_value);
          printf("  MPTCP aggregate delay benefit, within -1 and 1: %s\n", report->delay_value);
        }  

        terminate = 1;

        /* Initialize session variables */

        session_bytes = 0;
        session_nmetric = 0;

        break;
    }
  }

  get_now(&session_end, 0);
  if (gl.verbose) fprintf(stdout, "  Session total time : %.6f\n", time_to_seconds( &session_start, &session_end ));
  id_client = 0; // get ready for another client id
  dataport = gl.dataport; //ini_dataport;
}

void HandleSendingClient(int sock, u_int32_t nbytes) 
{
  u_int32_t tot_received = 0;

  tot_received = Receiver(DEV_SERVER, session_nmetric, sock, session_bytes, slices, NULL);
  
  printf("  Received: %u bytes\n", tot_received);  
}

void HandleReceivingClient(int sock, u_int32_t nbytes) 
{
  u_int32_t transmitted = 0;
 
  transmitted = Sender(DEV_SERVER, session_nmetric, sock, nbytes, session_timettx, session_cbrate, slices, NULL);

  printf("  Sent %d bytes.\n", transmitted);
}
